/* Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>

#if (WIN32|WINNT)
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_spell.h"
#include "udm_robots.h"
#include "udm_mutex.h"
#include "udm_db.h"
#include "udm_unicode.h"
#include "udm_url.h"
#include "udm_log.h"
#include "udm_proto.h"
#include "udm_conf.h"
#include "udm_crc32.h"
#include "udm_xmalloc.h"
#include "udm_cache.h"
#include "udm_boolean.h"
#include "udm_searchtool.h"
#include "udm_searchcache.h"
#include "udm_server.h"
#include "udm_stopwords.h"
#include "udm_doc.h"
#include "udm_result.h"
#include "udm_vars.h"
#include "udm_agent.h"
#include "udm_store.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_db_int.h"
#include "udm_sqldbms.h"
#include "udm_match.h"
#include "udm_indexer.h"

#if HAVE_ORACLE8
static void param_add(UDM_DB *db, int pos_out_1, int pos_out_2, int pos_out_3, int pos_out_4){
        db->par->out_pos_val[0][db->par->out_rec] = pos_out_1;
        db->par->out_pos_val[1][db->par->out_rec] = pos_out_2;
        db->par->out_pos_val[2][db->par->out_rec] = pos_out_3;
        db->par->out_pos_val[3][db->par->out_rec] = pos_out_4;
        db->par->out_rec++;
}

static void param_init(UDM_DB *db, int pos1, int pos2, int pos3, int pos4){
        db->par->out_pos[0]=pos1;
        db->par->out_pos[1]=pos2;
        db->par->out_pos[2]=pos3;
        db->par->out_pos[3]=pos4;
}

#endif

static void AutoWild(UDM_AGENT * Agent,UDM_VARLIST * vars){
	size_t i;
	
	for(i=0;i<vars->nvars;i++){
		if(!strcmp(vars->Var[i].name,"ul")){
			if((vars->Var[i].val)&&(vars->Var[i].val[0])){
				UDM_URL url;
				char * wild;
				
				wild=(char*)malloc(strlen(vars->Var[i].val)+3);
				UdmURLParse(&url,vars->Var[i].val);
				if(url.schema[0]){
					sprintf(wild,"%s%%",vars->Var[i].val);
				}else{
					sprintf(wild,"%%%s%%",vars->Var[i].val);
				}
				free(vars->Var[i].val);
				vars->Var[i].val=wild;
			}
		}
	}
}




static const char *BuildWhere(UDM_ENV * Conf,UDM_DB *db){
	size_t	i;
	char	*urlstr;
	char	*tagstr;
	char	*statusstr;
	char	*catstr;
	char	*langstr;
	char    *typestr;
	char    *fromstr;
	char    *serverstr;
	char    *sitestr;
	char    *timestr;
	int     fromserver = 1, fromurlinfo_lang = 1, fromurlinfo_type = 1;
	int     dt = UDM_DT_UNKNOWN, dx = 1, dm = 1, dy = 1970, dd = 1;
	time_t  dp = 0, DB = 0, DE = time(NULL);
	struct tm tm;
	
	if(db->where)return(db->where);
	
	bzero((void*)&tm, sizeof(struct tm));
	urlstr=strdup("");
	tagstr=strdup("");
	statusstr=strdup("");
	catstr=strdup("");
	langstr=strdup("");
	typestr = strdup("");
	fromstr = strdup("");
	serverstr = strdup("");
	sitestr = strdup("");
	timestr = strdup("");
	
	
	for(i=0;i<Conf->Vars.nvars;i++){
		const char *var=Conf->Vars.Var[i].name?Conf->Vars.Var[i].name:"";
		const char *val=Conf->Vars.Var[i].val?Conf->Vars.Var[i].val:"";
		int intval=atoi(val);
		
		if(!val[0])continue;
		
		if(!strcmp(var,"tag") || !strcmp(var,"t")){
			tagstr=(char*)realloc(tagstr,strlen(tagstr)+strlen(val)+50);
			if(tagstr[0])strcpy(UDM_STREND(tagstr)-1," OR ");
			else	strcat(tagstr,"(");

			if(db->DBType==UDM_DB_PGSQL)
				sprintf(UDM_STREND(tagstr),"(s.tag || '') LIKE '%s')",val);
			else
				sprintf(UDM_STREND(tagstr),"s.tag LIKE '%s')",val);
			if (fromserver) {
			  fromserver = 0;
			  fromstr = (char*)realloc(fromstr, strlen(fromstr) + 12);
			  sprintf(UDM_STREND(fromstr), ", server s");
			  serverstr = (char*)realloc(serverstr, strlen(serverstr) + 64);
			  sprintf(UDM_STREND(serverstr), " AND s.rec_id=url.server_id");
			}
		}
		
		if(!strcmp(var,"s")){
			statusstr=(char*)realloc(statusstr,strlen(statusstr)+strlen(val)+50);
			
			if(db->DBSQL_IN){
				if(statusstr[0])sprintf(UDM_STREND(statusstr)-1,",%d)",intval);
				else	sprintf(statusstr," url.status IN (%d)",intval);
			}else{
				if(statusstr[0])strcpy(UDM_STREND(statusstr)-1," OR ");
				else	strcat(statusstr,"(");
				sprintf(UDM_STREND(statusstr),"url.status=%d)",intval);
			}
		}

		if(!strcmp(var,"ul")){
			UDM_URL	URL;
			const char *first = "%";

			UdmURLParse(&URL,val);
			urlstr=(char*)realloc(urlstr,strlen(urlstr)+strlen(val)+50);

			if(URL.schema[0] && URL.hostinfo[0]){
			  first = "";
			}
			if(urlstr[0])strcpy(UDM_STREND(urlstr)-1," OR ");
			else	strcat(urlstr,"(");
			if(db->DBType==UDM_DB_PGSQL)
				sprintf(UDM_STREND(urlstr),"(url.url || '') LIKE '%s%s%%')", first, val);
			else
				sprintf(UDM_STREND(urlstr),"url.url LIKE '%s%s%%')", first, val);
		}
		
		if(!strcmp(var,"u")){
			urlstr=(char*)realloc(urlstr,strlen(urlstr)+strlen(val)+50);
			if(urlstr[0])strcpy(UDM_STREND(urlstr)-1," OR ");
			else	strcat(urlstr,"(");
			if(db->DBType==UDM_DB_PGSQL)
				sprintf(UDM_STREND(urlstr),"(url.url || '') LIKE '%s')", val);
			else
				sprintf(UDM_STREND(urlstr),"url.url LIKE '%s')", val);
		}
		
		if(!strcmp(var,"lang") || !strcmp(var,"g")){
			langstr=(char*)realloc(langstr,strlen(langstr)+strlen(val)+50);
			if(langstr[0])strcpy(UDM_STREND(langstr)-1," OR ");
			else	strcat(langstr,"(");
			sprintf(UDM_STREND(langstr),"il.sval LIKE '%s')",val);
			if (fromurlinfo_lang) {
			  fromurlinfo_lang = 0;
			  fromstr = (char*)realloc(fromstr, strlen(fromstr) + 16);
			  sprintf(UDM_STREND(fromstr), ", urlinfo il");
			  serverstr = (char*)realloc(serverstr, strlen(serverstr) + 64);
			  sprintf(UDM_STREND(serverstr), " AND il.url_id=url.rec_id AND il.sname='Content-Language'");
			}
		}
		
		if(!strcmp(var,"cat")){
			catstr=(char*)realloc(catstr,strlen(catstr)+strlen(val)+50);
			if(catstr[0])strcpy(UDM_STREND(catstr)-1," OR ");
			else	strcat(catstr,"(");
			sprintf(UDM_STREND(catstr),"s.category LIKE '%s%%')",val);
			if (fromserver) {
			  fromserver = 0;
			  fromstr = (char*)realloc(fromstr, strlen(fromstr) + 12);
			  sprintf(UDM_STREND(fromstr), ", server s");
			  serverstr = (char*)realloc(serverstr, strlen(serverstr) + 64);
			  sprintf(UDM_STREND(serverstr), " AND s.rec_id=url.server_id");
			}
		}
		if(!strcmp(var,"type")){
			typestr = (char*)realloc(typestr, strlen(typestr) + strlen(val) + 50);
			if(typestr[0]) strcpy(UDM_STREND(typestr) - 1, " OR ");
			else	strcat(typestr,"(");

			if(db->DBType==UDM_DB_PGSQL)
				sprintf(UDM_STREND(typestr),"(it.sval || '') LIKE '%s')",val);
			else
				sprintf(UDM_STREND(typestr),"it.sval LIKE '%s')",val);
			if (fromurlinfo_type) {
			  fromurlinfo_type = 0;
			  fromstr = (char*)realloc(fromstr, strlen(fromstr) + 16);
			  sprintf(UDM_STREND(fromstr), ", urlinfo it");
			  serverstr = (char*)realloc(serverstr, strlen(serverstr) + 64);
			  sprintf(UDM_STREND(serverstr), " AND it.url_id=url.rec_id AND it.sname='Content-Type'");
			}
		}
		if(!strcmp(var,"site") && intval > 0) {
			sitestr=(char*)realloc(sitestr, strlen(sitestr) + strlen(val) + 50);
			
			if(db->DBSQL_IN){
				if (sitestr[0]) sprintf(UDM_STREND(sitestr) - 1, ",%d)", intval);
				else	sprintf(sitestr, " url.site_id IN (%d)", intval);
			}else{
				if (sitestr[0]) strcpy(UDM_STREND(sitestr) - 1, " OR ");
				else	strcat(sitestr, "(");
				sprintf(UDM_STREND(sitestr), "url.site_id=%d)", intval);
			}
		}
		if (!strcmp(var, "dt")) {
		  if(!strcasecmp(val, "back")) dt = UDM_DT_BACK;
		  else if (!strcasecmp(val, "er")) dt = UDM_DT_ER;
		  else if (!strcasecmp(val, "range")) dt = UDM_DT_RANGE;
		}
		if (!strcmp(var, "dx")) {
		  if (intval == 1 || intval == -1) dx = intval;
		  else dx = 1;
		}
		if (!strcmp(var, "dm")) {
		  dm = (intval) ? intval : 1;
		}
		if (!strcmp(var, "dy")) {
		  dy = (intval) ? intval : 1970;
		}
		if (!strcmp(var, "dd")) {
		  dd = (intval) ? intval : 1;
		}
		if (!strcmp(var, "dp")) {
		  dp = Udm_dp2time_t(val);
		}
		if (!strcmp(var, "db")) {
		  sscanf(val, "%d/%d/%d", &tm.tm_mday, &tm.tm_mon, &tm.tm_year);
		  tm.tm_year -= 1900;
		  DB = mktime(&tm);
		}
		if (!strcmp(var, "de")) {
		  sscanf(val, "%d/%d/%d", &tm.tm_mday, &tm.tm_mon, &tm.tm_year);
		  tm.tm_year -= 1900;
		  DE = mktime(&tm);
		}
	}
	
	switch(dt) {
	case UDM_DT_BACK:
	  timestr = (char*)realloc(timestr, 128);
	  if (dp) sprintf(timestr, "url.last_mod_time >= %li", time(NULL) - dp);
	  break;
	case UDM_DT_ER:
	  timestr = (char*)realloc(timestr, 128);
	  tm.tm_mday = dd; tm.tm_mon = dm, tm.tm_year = dy - 1900;
	  sprintf(timestr, "url.last_mod_time %s %li", (dx == 1) ? ">=" : "<=", mktime(&tm));
	  break;
	case UDM_DT_RANGE:
	  timestr = (char*)realloc(timestr, 128);
	  sprintf(timestr, "url.last_mod_time >= %li AND url.last_mod_time <= %li", DB, DE);
	  break;
	case UDM_DT_UNKNOWN:
	default:
	  break;
	}


	if( !urlstr[0] && !tagstr[0] && !statusstr[0] && !catstr[0] && !langstr[0] 
	    && !typestr[0] && !serverstr[0] && !fromstr[0] && !sitestr[0] && !timestr[0] ){
		db->where=strdup("");
		db->from = strdup("");
		goto ret;
	}
	i=strlen(urlstr)+strlen(tagstr)+strlen(statusstr) + strlen(catstr) + strlen(langstr) 
	  + strlen(typestr) + strlen(serverstr) + strlen(sitestr) + strlen(timestr);
	db->where=(char*)malloc(i+100);
	db->where[0] = '\0';
	db->from = strdup(fromstr);
	
	if(urlstr[0]){
		strcat(db->where,urlstr);
	}
	if(tagstr[0]){
		if(db->where[0])strcat(db->where," AND ");
		strcat(db->where,tagstr);
	}
	if(statusstr[0]){
		if(db->where[0])strcat(db->where," AND ");
		strcat(db->where,statusstr);
	}
	if(catstr[0]){
		if(db->where[0])strcat(db->where," AND ");
		strcat(db->where,catstr);
	}
	if(langstr[0]){
		if(db->where[0])strcat(db->where," AND ");
		strcat(db->where,langstr);
	}
	if(typestr[0]){
		if(db->where[0]) strcat(db->where, " AND ");
		strcat(db->where, typestr);
	}
	if(sitestr[0]){
		if(db->where[0]) strcat(db->where, " AND ");
		strcat(db->where, sitestr);
	}
	if(serverstr[0]){
		if(!db->where[0]) strcat(db->where, " 1 ");
		strcat(db->where, serverstr);
	}
	if(timestr[0]){
		if(db->where[0]) strcat(db->where, " AND ");
		strcat(db->where, timestr);
	}
ret:
	free(urlstr);
	free(tagstr);
	free(statusstr);
	free(catstr);
	free(langstr);
	free(typestr);
	free(fromstr);
	free(serverstr);
	free(sitestr);
	free(timestr);
	return db->where;
}


/********************************************************************/
/* --------------------------------------------------------------- */
/* Almost Unified code for all supported SQL backends              */
/* --------------------------------------------------------------- */
/*******************************************************************/


static const int search_cache_size=1000;


/************* Servers ******************************************/

static char * strdupnull(const char * src){
	if(src){
		if(*src){
			return(strdup(src));
		}else{
			return strdup("");
		}
	}else{
		return strdup("");
	}
}

static int UdmLoadServerTable(UDM_AGENT * Indexer, UDM_SERVERLIST *S, UDM_DB *db){
	size_t		rows,i;
	UDM_SQLRES	SQLRes;
	UDM_HREF	Href;
	char		qbuf[1024];
	const char	*name=db->addr.filename[0]?db->addr.filename:"server";
	int		res;
	
	snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT rec_id,url,alias,period,tag,category,charset,lang,basic_auth,\
proxy,proxy_auth,\
maxhops,gindex,follow,use_robots,use_clones,\
max_net_errors,net_delay_time,read_timeout,\
match_type,ordre,command,weight,case_sense,nomatch \
FROM %s WHERE enabled=1 AND parent=0 ORDER BY ordre",name);
	
	if(UDM_OK!=(res=UdmSQLQuery(db,&SQLRes,qbuf)))
		return res;
	
	bzero((void*)&Href, sizeof(Href));
	
	rows=UdmSQLNumRows(&SQLRes);
	for(i=0;i<rows;i++){
		const char	*val;
		UDM_SERVER	Server;
		
		UdmServerInit(&Server);
		
		Server.site_id		= UDM_ATOI(UdmSQLValue(&SQLRes, i, 0));
		Server.Match.pattern	= strdupnull(UdmSQLValue(&SQLRes,i,1));
		Server.Match.match_type	= UDM_ATOI(UdmSQLValue(&SQLRes,i,19));
		Server.ordre		= UDM_ATOI(UdmSQLValue(&SQLRes,i,20));
		Server.command		= *UdmSQLValue(&SQLRes,i,21);
		Server.weight		= UDM_ATOI(UdmSQLValue(&SQLRes,i,22));
		Server.Match.case_sense	= UDM_ATOI(UdmSQLValue(&SQLRes,i,23));
		Server.Match.nomatch	= UDM_ATOI(UdmSQLValue(&SQLRes,i,24));
		
		if((val=UdmSQLValue(&SQLRes,i,2)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Alias",val);
		
		if((val=UdmSQLValue(&SQLRes,i,3)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Period",val);
		
		if((val=UdmSQLValue(&SQLRes,i,4)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Tag",val);
		
		if((val=UdmSQLValue(&SQLRes,i,5)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Category",val);
		
		if((val=UdmSQLValue(&SQLRes,i,6)) && val[0])
			UdmVarListAddStr(&Server.Vars,"RemoteCharset",val);
		
		if((val=UdmSQLValue(&SQLRes,i,7)) && val[0])
			UdmVarListAddStr(&Server.Vars,"DefaultLang",val);
		
		if((val=UdmSQLValue(&SQLRes,i,8)) && val[0]){
			char	*basic_auth;
			basic_auth=(char*)malloc(BASE64_LEN(strlen(val)));
			udm_base64_encode(val,basic_auth,strlen(val));
			UdmVarListAddStr(&Server.Vars,"Request.Authorization",basic_auth);
			UDM_FREE(basic_auth);
		}
		
		if((val=UdmSQLValue(&SQLRes,i,9)) && val[0]){
			UdmVarListAddStr(&Server.Vars,"Request.Proxy",val);/* FIXME: separate port */
		}
		
		if((val=UdmSQLValue(&SQLRes,i,10)) && val[0]){
			char	*basic_auth;
			basic_auth=(char*)malloc(BASE64_LEN(strlen(val)));
			udm_base64_encode(val,basic_auth,strlen(val));
			UdmVarListAddStr(&Server.Vars,"Request.Proxy-Authorization",basic_auth);
			UDM_FREE(basic_auth);
		}
		
		if((val=UdmSQLValue(&SQLRes,i,11)) && val[0])
			UdmVarListAddStr(&Server.Vars,"MaxHops",val);
		
		if((val=UdmSQLValue(&SQLRes,i,12)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Index",val);
		
		if((val=UdmSQLValue(&SQLRes,i,13)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Follow",val);
		
		if((val=UdmSQLValue(&SQLRes,i,14)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Robots",val);
		
		if((val=UdmSQLValue(&SQLRes,i,15)) && val[0])
			UdmVarListAddStr(&Server.Vars,"Clones",val);
		
		if((val=UdmSQLValue(&SQLRes,i,16)) && val[0])
			UdmVarListAddStr(&Server.Vars,"MaxNetErrors",val);
		
		if((val=UdmSQLValue(&SQLRes,i,17)) && val[0])
			UdmVarListAddStr(&Server.Vars,"NetErrorDelayTime",val);
		
		if((val=UdmSQLValue(&SQLRes,i,18)) && val[0])
			UdmVarListAddStr(&Server.Vars,"ReadTimeOut",val);
		
		if (*UdmSQLValue(&SQLRes, i, 21) == 'S') {
			UdmServerAdd(Indexer->Conf,&Server);
			if(Indexer->flags & UDM_FLAG_ADD_SERV) {
				Href.url=Server.Match.pattern;
				Href.method=UDM_METHOD_GET;
				Href.site_id = Server.site_id;
				Href.server_id = Server.site_id;
				UdmHrefListAdd(&Indexer->Conf->Hrefs,&Href);
			  }
		} else {
			char errstr[128];
			UdmMatchListAdd(&Indexer->Conf->Filters,&Server.Match,errstr,sizeof(errstr));
		}
		
		UdmServerFree(&Server);
	}
	UdmSQLFree(&SQLRes);
	return(UDM_OK);
}

static int UdmServerTableFlush(UDM_DB *db){
	UDM_SQLRES	SQLRes;
	int rc;

	rc = UdmSQLQuery(db, &SQLRes, "UPDATE server SET enabled=0 WHERE parent=0");
	UdmSQLFree(&SQLRes);
	return rc;
}

static int UdmServerTableAdd(UDM_SERVERLIST *S, UDM_DB *db) {
	UDM_SQLRES	SQLRes;
	int		res=UDM_OK;
	const char	*alias=UdmVarListFindStr(&S->Server->Vars,"Alias",NULL);
	size_t		len;
	char		*buf;
	const char	*id_field;
	const char	*id_value;
	
	len = ((S->Server->Match.pattern)?strlen(S->Server->Match.pattern):0) + ((alias)?strlen(alias):0) + 2048;
	buf = (char*)malloc(len);
	if (buf == NULL) {
		strcpy(db->errstr, "Out of memory");
		db->errcode = 1;
		return UDM_ERROR;
	}
	
	switch(db->DBType){
		case UDM_DB_IBASE:
			id_field="rec_id,";
			id_value="GEN_ID(rec_id_GEN,1),";
			break;
		default:;
			id_field="";
			id_value="";
	}
	
	snprintf(buf, len, "\
INSERT INTO server (%s\
enabled, alias, period, tag, category, charset, lang, \
basic_auth, proxy, proxy_auth,\
maxhops, gindex, follow, use_robots, use_clones, \
max_net_errors, net_delay_time, read_timeout, \
match_type,case_sense,nomatch,command, parent, ordre, weight, url \
) VALUES (%s\
1, '%s', %d, '%s', '%s', '%s', '%s', '%s', '%s', '%s', %d, %d, %d, %d,\
%d, %d, %d, %d, %d, %d, %d, '%c', %d, %d, %d, '%s'\
)",
		 id_field,
		 id_value,
		 (alias)?(alias):"",
		 UdmVarListFindInt(&S->Server->Vars, "Period", UDM_DEFAULT_REINDEX_TIME),
		 UdmVarListFindStr(&S->Server->Vars, "Tag", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Category", ""),
		 UdmVarListFindStr(&S->Server->Vars, "RemoteCharset", ""),
		 UdmVarListFindStr(&S->Server->Vars, "DefaultLang", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Authorization", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Proxy", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Proxy-Authorization", ""),
		 UdmVarListFindInt(&S->Server->Vars, "MaxHops", UDM_DEFAULT_MAX_HOPS),
		 UdmVarListFindInt(&S->Server->Vars, "Index", 1),
		 UdmVarListFindInt(&S->Server->Vars, "Follow", UDM_FOLLOW_PATH),
		 UdmVarListFindInt(&S->Server->Vars, "Robots", 1),
		 UdmVarListFindInt(&S->Server->Vars, "Clones", 1),
		 UdmVarListFindInt(&S->Server->Vars, "MaxNetErrors", UDM_MAXNETERRORS),
		 UdmVarListFindInt(&S->Server->Vars, "NetErrorDelayTime", UDM_DEFAULT_NET_ERROR_DELAY_TIME),
		 UdmVarListFindInt(&S->Server->Vars, "ReadTimeOut", UDM_READ_TIMEOUT),
		 S->Server->Match.match_type,
		 S->Server->Match.case_sense,
		 S->Server->Match.nomatch,
		 S->Server->command,
		 S->Server->parent,
		 S->Server->ordre,
		 S->Server->weight,
		 S->Server->Match.pattern
		 );
	
	if(UDM_OK!= (res = UdmSQLQuery(db, &SQLRes, buf)))
		goto ex;
	
	snprintf(buf, len, "\
UPDATE server SET enabled=1, alias='%s', period=%d, tag='%s', category='%s', charset='%s', lang='%s',\
basic_auth='%s', proxy='%s',proxy_auth='%s', maxhops=%d, gindex=%d, follow=%d,\
use_robots=%d, use_clones=%d, max_net_errors=%d, net_delay_time=%d,read_timeout=%d,\
match_type=%d,case_sense=%d,nomatch=%d,command='%c', parent=%d, ordre=%d, weight=%d \
WHERE url='%s'",
		 (alias)?(alias):"",
		 UdmVarListFindInt(&S->Server->Vars, "Period", UDM_DEFAULT_REINDEX_TIME),
		 UdmVarListFindStr(&S->Server->Vars, "Tag", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Category", ""),
		 UdmVarListFindStr(&S->Server->Vars, "RemoteCharset", ""),
		 UdmVarListFindStr(&S->Server->Vars, "DefaultLang", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Authorization", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Proxy", ""),
		 UdmVarListFindStr(&S->Server->Vars, "Proxy-Authorization", ""),
		 UdmVarListFindInt(&S->Server->Vars, "MaxHops", UDM_DEFAULT_MAX_HOPS),
		 UdmVarListFindInt(&S->Server->Vars, "Index", 1),
		 UdmVarListFindInt(&S->Server->Vars, "Follow", UDM_FOLLOW_PATH),
		 UdmVarListFindInt(&S->Server->Vars, "Robots", 1),
		 UdmVarListFindInt(&S->Server->Vars, "Clones", 1),
		 UdmVarListFindInt(&S->Server->Vars, "MaxNetErrors", UDM_MAXNETERRORS),
		 UdmVarListFindInt(&S->Server->Vars, "NetErrorDelayTime", UDM_DEFAULT_NET_ERROR_DELAY_TIME),
		 UdmVarListFindInt(&S->Server->Vars, "ReadTimeOut", UDM_READ_TIMEOUT),
		 S->Server->Match.match_type,
		 S->Server->Match.case_sense,
		 S->Server->Match.nomatch,
		 S->Server->command,
		 S->Server->parent,
		 S->Server->ordre,
		 S->Server->weight,
		 S->Server->Match.pattern
		 );
	
	if(UDM_OK != (res = UdmSQLQuery(db, &SQLRes, buf)))
		goto ex;
	
	snprintf(buf, len, "SELECT rec_id FROM server WHERE url='%s' AND command='%c'",
		 S->Server->Match.pattern, S->Server->command);
	
	if (UDM_OK != (res = UdmSQLQuery(db, &SQLRes, buf)))
		goto ex;
	
	S->Server->site_id = UDM_ATOI(UdmSQLValue(&SQLRes,0,0));
	UdmSQLFree(&SQLRes);
ex:
	free(buf);
	return res;
}

static int UdmServerTableGetId(UDM_AGENT *Indexer, UDM_SERVERLIST *S, UDM_DB *db) {
	UDM_SQLRES	SQLRes;
	size_t len = ((S->Server->Match.pattern)?strlen(S->Server->Match.pattern):0) + 1024;
	char *buf = (char*)malloc(len);
	int res, id = 0, i;
	const char	*id_field;
	const char	*id_value;
	

	if (buf == NULL) {
	  UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
	  return UDM_ERROR;
	}

	for (i = 0; i < UDM_SERVERID_CACHE_SIZE; i++) {
	  if (Indexer->ServerIdCacheCommand[i] == S->Server->command)
	    if (!strcmp(Indexer->ServerIdCache[i], S->Server->Match.pattern)) {
	      id = Indexer->ServerIdCacheId[i];
	      break;
	    }
	}

	if (id == 0) {
	
	  snprintf(buf, len, "SELECT rec_id FROM server WHERE command='%c' AND url='%s'",
		 S->Server->command,
		 S->Server->Match.pattern
		 );
	  res = UdmSQLQuery(db, &SQLRes, buf);
	  if ((res == UDM_OK) && UdmSQLNumRows(&SQLRes)) {
	    id = UDM_ATOI(UdmSQLValue(&SQLRes, 0, 0));
	    UDM_FREE(Indexer->ServerIdCache[Indexer->pServerIdCache]);
	    Indexer->ServerIdCache[Indexer->pServerIdCache]=strdup(S->Server->Match.pattern);
	    Indexer->ServerIdCacheCommand[Indexer->pServerIdCache] = S->Server->command;
	    Indexer->ServerIdCacheId[Indexer->pServerIdCache] = id;
	    Indexer->pServerIdCache = (Indexer->pServerIdCache + 1) % UDM_SERVERID_CACHE_SIZE;
	    free(buf);
	    UdmSQLFree(&SQLRes);
	    return id;
	  }
	  snprintf(buf, len, "SELECT enabled, alias, period, tag, category, charset, lang, basic_auth, proxy, proxy_auth,\
maxhops, gindex, follow, use_robots, use_clones, max_net_errors, net_delay_time, read_timeout, match_type,\
command, ordre, url, weight FROM server WHERE rec_id=%d", S->Server->parent);
	  res = UdmSQLQuery(db, &SQLRes, buf);
	  if (res != UDM_OK) {
	    free(buf);
	    UdmSQLFree(&SQLRes);
	    return 0;
	  }

	switch(db->DBType){
		case UDM_DB_IBASE:
			id_field="rec_id,";
			id_value="GEN_ID(rec_id_GEN,1),";
			break;
		default:;
			id_field="";
			id_value="";
	}
	
	  snprintf(buf, len, "\
INSERT INTO server (%s\
enabled, alias, period, tag, category, charset, lang, basic_auth, proxy, proxy_auth,\
maxhops, gindex, follow, use_robots, use_clones, max_net_errors, net_delay_time, read_timeout, match_type,\
command, parent, ordre, weight, url) \
VALUES (%s\
%d, '%s', %d, '%s', '%s', '%s', '%s', '%s', '%s', '%s', %d, %d, %d, %d,\
%d, %d, %d, %d, %d, '%c', %d, %d, %d, '%s')\
",
		 id_field,
		 id_value,
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 0)),
		 UdmSQLValue(&SQLRes, 0, 1),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 2)), 
		 UdmSQLValue(&SQLRes, 0, 3), 
		 UdmSQLValue(&SQLRes, 0, 4),
		 UdmSQLValue(&SQLRes, 0, 5), 
		 UdmSQLValue(&SQLRes, 0, 6),
		 UdmSQLValue(&SQLRes, 0, 7),
		 UdmSQLValue(&SQLRes, 0, 8),
		 UdmSQLValue(&SQLRes, 0, 9),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 10)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 11)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 12)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 13)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 14)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 15)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 16)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 17)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 18)),
		 S->Server->command,
		 S->Server->parent,
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 20)),
		 UDM_ATOI(UdmSQLValue(&SQLRes, 0, 22)),
		 S->Server->Match.pattern
		 );
	  res = UdmSQLQuery(db, &SQLRes, buf);

	  snprintf(buf, len, "SELECT rec_id FROM server WHERE url='%s' AND command='%c' AND parent > 0",
		 S->Server->Match.pattern,
		 S->Server->command);
		 
	  res = UdmSQLQuery(db, &SQLRes, buf);
	  if (res == UDM_OK) {
	    id = UDM_ATOI(UdmSQLValue(&SQLRes,0,0));
	    UDM_FREE(Indexer->ServerIdCache[Indexer->pServerIdCache]);
	    Indexer->ServerIdCache[Indexer->pServerIdCache] = strdup(S->Server->Match.pattern);
	    Indexer->ServerIdCacheCommand[Indexer->pServerIdCache] = S->Server->command;
	    Indexer->ServerIdCacheId[Indexer->pServerIdCache] = id;
	    Indexer->pServerIdCache = (Indexer->pServerIdCache + 1) % UDM_SERVERID_CACHE_SIZE;
	  }
	  UdmSQLFree(&SQLRes);
	}
	free(buf);
	return id;
}

/************************* find url ********************************/

static int UdmFindURL(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc, UDM_DB *db){
	UDM_SQLRES	SQLRes;
	const char	*url=UdmVarListFindStr(&Doc->Sections,"URL","");
	int		id=0;
	int		use_crc32_url_id=!strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"UseCRC32URLId","no"),"yes");
	int		rc=UDM_OK;
	
	if(use_crc32_url_id){
		/* Auto generation of rec_id */
		/* using CRC32 algorythm     */
		id=UdmStrCRC32(url);
	}else{
		size_t i, l = 127;
		const char *o;
		char *qbuf = NULL;
		char *e_url = NULL;
		
		/* Escape URL string */
		if ((e_url = (char*)malloc( l = (8 * strlen(url)) )) == NULL) {
		  UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
		  return UDM_ERROR;
		}
		if ((qbuf = (char*)malloc( l + 100 )) == NULL){
			UDM_FREE(e_url);
			UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
			return UDM_ERROR;
		}
		UdmDBEscStr(db->DBType,e_url,url,strlen(url));


		for(i = 0; i < UDM_FINDURL_CACHE_SIZE; i++) {
		  if (Indexer->UdmFindURLCache[i])
		    if (!strcmp(e_url, Indexer->UdmFindURLCache[i])) {
		      id = Indexer->UdmFindURLCacheId[i];
		      break;
		    }
		}

		if (id == 0) {
		  snprintf(qbuf, l + 100, "SELECT rec_id FROM url WHERE url='%s'",e_url);
		  if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes,qbuf))){
			UDM_FREE(e_url);
			UDM_FREE(qbuf);
			return rc;
		  }
		  for(i=0;i<UdmSQLNumRows(&SQLRes);i++){
			if((o=UdmSQLValue(&SQLRes,i,0))){
				id=atoi(o);
				break;
			}
		  }
		  UdmSQLFree(&SQLRes);
		  UDM_FREE(Indexer->UdmFindURLCache[Indexer->pURLCache]);
		  Indexer->UdmFindURLCache[Indexer->pURLCache] = strdup(e_url);
		  Indexer->UdmFindURLCacheId[Indexer->pURLCache] = id;
		  Indexer->pURLCache = (Indexer->pURLCache + 1) % UDM_FINDURL_CACHE_SIZE;
		}
		UDM_FREE(e_url);
		UDM_FREE(qbuf);
	}
	UdmVarListReplaceInt(&Doc->Sections,"ID",id);
	return	rc;
}


static int UdmFindMessage(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc, UDM_DB *db){
	size_t 		i;
	char 		qbuf[1024];
	char 		eid[UDM_URLSIZE*4];
	UDM_SQLRES	SQLRes;
	const char	*message_id=UdmVarListFindStr(&Doc->Sections,"Header.Message-ID",NULL);
	int		rc;
	
	if(!message_id)
		return UDM_OK;
	
	/* Escape URL string */
	UdmDBEscStr(db->DBType,eid,message_id,strlen(message_id));
	
	snprintf(qbuf, 1023, "SELECT rec_id FROM url u, urlinfo i WHERE u.rec_id=i.url_id AND i.sname='Message-ID' AND i.sval='%s'", eid);
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes,qbuf)))
		return rc;
	
	for(i=0;i<UdmSQLNumRows(&SQLRes);i++){
		const char * o;
		if((o=UdmSQLValue(&SQLRes,i,0))){
			UdmVarListReplaceInt(&Doc->Sections,"ID",atoi(o));
			break;
		}
	}
	UdmSQLFree(&SQLRes);
	return(UDM_OK);
}


/********************** Words ***********************************/

static int UdmDeleteWordFromURL(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	char	qbuf[512];
	int	i = 0, rc = UDM_OK;
	size_t  last = 0;
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	
	switch(db->DBMode){
	
	case UDM_DBMODE_MULTI:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				snprintf(qbuf,sizeof(qbuf),"DELETE FROM dict%d WHERE url_id=%d",DICTNUM(i),url_id);
				if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
					return rc;
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_MULTI_CRC:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				int done=0;
#ifdef HAVE_ORACLE8
				if(db->DBDriver==UDM_DB_ORACLE8){
					snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict%d WHERE url_id=:1", DICTNUM(i),url_id);
					param_init(db, 1, 0, 0, 0);
					param_add(db, url_id, 0, 0, 0);
					done=1;
				}
#endif				
				if(!done)
					snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict%d WHERE url_id=%d",DICTNUM(i),url_id);

				if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
					return rc;
				
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_SINGLE_CRC:
		snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict WHERE url_id=%d",url_id);
		if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
			return rc;
		break;
	case UDM_DBMODE_BLOB:
		/*i=UdmBlobWriteStamp(Indexer,Doc);*/
		return(i);
	case UDM_DBMODE_CACHE:
		/* Let's lock it */
		i=UdmDeleteURLFromCache(Indexer,url_id,db);
		return(i);
	default:  /* UDM_DBMODE_SINGLE */
		snprintf(qbuf, sizeof(qbuf)-1, "DELETE FROM dict WHERE url_id=%d", url_id);
		if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
			return(UDM_ERROR);
		break;
	}
	return(UDM_OK);
}

static int StoreWordsMulti(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_DB *db){
	char	qbuf[512];
	char	tablename[64]="dict";
	char	tbl_nm[64];
	int	n, rc = UDM_OK;
	size_t  prev_dictlen = 0;
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	
	if(db->DBMode==UDM_DBMODE_MULTI){
		strcpy(tablename,"dict");
	}else{
		strcpy(tablename,"ndict");
	}
	
	for(n=0;n<NDICTS;n++){
		if(prev_dictlen==dictlen[n])continue;
		prev_dictlen=dictlen[n];

		sprintf(tbl_nm, "%s%d", tablename, dictlen[n]);
		if(1){
			switch(db->DBType){
				case UDM_DB_PGSQL:
					rc=UdmSQLQuery(db,NULL,"BEGIN WORK");
					break;
				case UDM_DB_ORACLE7:
				case UDM_DB_ORACLE8:
				case UDM_DB_SAPDB:
					rc=UdmSQLQuery(db,NULL,"COMMIT");
					db->commit_fl = 1;
					break;
				default:
					db->commit_fl = 1;
					break;
			}
			if(rc!=UDM_OK)
				return rc;
		}
		
		/* Delete old words */
		snprintf(qbuf,sizeof(qbuf),"DELETE FROM %s WHERE url_id=%d",tbl_nm,url_id);
		if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
			goto unlock_StoreWordsMulti;
		
		/* Insert new word */
		if(db->DBType==UDM_DB_MYSQL){
			int have_words=0;
			char * qb,*qe;
			size_t step=1024,mlen=1024,len,i;

			qb=(char*)malloc(mlen);
			if(db->DBMode==UDM_DBMODE_MULTI){
				sprintf(qb,"INSERT INTO %s (url_id,word,intag) VALUES ",tbl_nm);
			}else{
				sprintf(qb,"INSERT INTO %s (url_id,word_id,intag) VALUES ",tbl_nm);
			}
			qe=qb+strlen(qb);

			for(i=0;i<Doc->Words.nwords;i++){
				if(!Doc->Words.Word[i].coord)continue;
				
				if (DICTNUM(strlen(Doc->Words.Word[i].word)) == dictlen[n]){
					len=qe-qb;
					/* UDM_MAXWORDSIZE+100 should be enough */
					if((len+UDM_MAXWORDSIZE+100)>=mlen){
						mlen+=step;
						qb=(char*)realloc(qb,mlen);
						qe=qb+len;
					}
					if(have_words)strcpy(qe++,",");
					have_words++;
					if(db->DBMode==UDM_DBMODE_MULTI){
						sprintf(qe,"(%d,'%s',%d)",url_id,Doc->Words.Word[i].word,Doc->Words.Word[i].coord);
					}else{
						sprintf(qe,"(%d,%d,%d)",url_id,UdmStrCRC32(Doc->Words.Word[i].word),Doc->Words.Word[i].coord);
					}
					qe=qe+strlen(qe);
					
					/* Insert 64k packets max to stay */
					/* under MySQL's default limit. */
					if((qe-qb)>=64*1024){
						if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qb))){
							free(qb);
							goto unlock_StoreWordsMulti;
						}
						/* Init query again */
						if(db->DBMode==UDM_DBMODE_MULTI){
							sprintf(qb,"INSERT INTO %s (url_id,word,intag) VALUES ",tbl_nm);
						}else{
							sprintf(qb,"INSERT INTO %s (url_id,word_id,intag) VALUES ",tbl_nm);
						}
						qe=qb+strlen(qb);
						have_words = 0;
					}
				}
			}
			if(have_words)rc=UdmSQLQuery(db,NULL,qb);
			free(qb);
			if(rc!=UDM_OK) goto unlock_StoreWordsMulti;
		}else{
			size_t i;
			for(i=0;i<Doc->Words.nwords;i++){
				int len;
				if(!Doc->Words.Word[i].coord)continue;
				
				len=strlen(Doc->Words.Word[i].word);
				if (DICTNUM(len) == dictlen[n]){
					if(db->DBMode==UDM_DBMODE_MULTI){
						snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO %s (url_id,word,intag) VALUES(%d,'%s',%d)",tbl_nm,url_id,Doc->Words.Word[i].word,Doc->Words.Word[i].coord);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO %s (url_id,word_id,intag) VALUES(%d,%d,%d)",tbl_nm,url_id,UdmStrCRC32(Doc->Words.Word[i].word),Doc->Words.Word[i].coord);
					}
					if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
						goto unlock_StoreWordsMulti;
				}
			}
		}

unlock_StoreWordsMulti:		
		if(1){
			switch(db->DBType){
				case UDM_DB_PGSQL:
					rc=UdmSQLQuery(db,NULL,"END WORK");
					break;
				case UDM_DB_ORACLE7:
				case UDM_DB_ORACLE8:
				case UDM_DB_SAPDB:
					rc=UdmSQLQuery(db,NULL,"COMMIT");
					db->commit_fl = 0;
					break;
				default:
#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB || HAVE_DB2)
					rc=UdmSQLQuery(db,NULL,"UDM_COMMIT");
#endif
					db->commit_fl = 0;
					break;
			}
			if(rc!=UDM_OK)return rc;
		}
	}
	return(UDM_OK);
}


static int StoreWordsSingle(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_DB *db){
	size_t	i;
	char	qbuf[256]="";
	time_t	stmp;
	int	rc=UDM_OK;
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	
	stmp=time(NULL);
	
	/* Start transaction if supported */
	/* This is to make stuff faster   */

	if(1){
		switch(db->DBType){
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"BEGIN WORK");
				break;
			case UDM_DB_MSSQL:
				rc=UdmSQLQuery(db,NULL,"BEGIN TRANSACTION");
				break;
			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
			case UDM_DB_SAPDB:
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				db->commit_fl = 1;
				break;
			case UDM_DB_IBASE:
				rc=UdmSQLQuery(db,NULL,"BEGIN");
				db->commit_fl = 1;
				break;
			default:
				rc=UDM_OK;
				db->commit_fl = 1;
				break;
		}
		if(rc!=UDM_OK)return rc;
	}
	
	/* Delete old words */
	if(db->DBMode==UDM_DBMODE_SINGLE){
		sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	}else
	if(db->DBMode==UDM_DBMODE_SINGLE_CRC){
		sprintf(qbuf,"DELETE FROM ndict WHERE url_id=%d",url_id);
	}
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
		goto unlock_StoreWordsSingle;
	
	/* Insert new words */
	if(db->DBType==UDM_DB_MYSQL){
		if(Doc->Words.nwords){
			size_t nstored=0;
			
			while(nstored<Doc->Words.nwords){
				char * qb,*qe;
				size_t step=1024;
				size_t mlen=1024;

				qb=(char*)malloc(mlen);
				if(db->DBMode==UDM_DBMODE_SINGLE){
					strcpy(qb,"INSERT INTO dict (word,url_id,intag) VALUES ");
				}else
				if(db->DBMode==UDM_DBMODE_SINGLE_CRC){
					strcpy(qb,"INSERT INTO ndict (url_id,word_id,intag) VALUES ");
				}
				qe=qb+strlen(qb);

				for(i=nstored;i<Doc->Words.nwords;i++){
					size_t len=qe-qb;
					if(!Doc->Words.Word[i].coord) { nstored++; continue;}

				
					/* UDM_MAXWORDSIZE+100 should be enough */
					if((len+UDM_MAXWORDSIZE+100)>=mlen){
						mlen+=step;
						qb=(char*)realloc(qb,mlen);
						qe=qb+len;
					}
					
					if(i>nstored)*qe++=',';

					if(db->DBMode==UDM_DBMODE_SINGLE){
						*qe++='(';
						*qe++='\'';
						strcpy(qe,Doc->Words.Word[i].word);
						while(*qe)qe++;
						*qe++='\'';
						*qe++=',';
						qe+=sprintf(qe,"%d,%d",url_id,Doc->Words.Word[i].coord);
						*qe++=')';
						*qe='\0';
					}else
					if(db->DBMode==UDM_DBMODE_SINGLE_CRC){
						sprintf(qe,"(%d,%d,%d)",url_id,UdmStrCRC32(Doc->Words.Word[i].word),Doc->Words.Word[i].coord);
						qe=qe+strlen(qe);
					}else
					if(db->DBMode==UDM_DBMODE_BLOB){
						sprintf(qe,"('%s',%d,%d,%li)",Doc->Words.Word[i].word,url_id,Doc->Words.Word[i].coord,stmp);
						qe=qe+strlen(qe);
					}
					if(qe>qb+UDM_MAX_MULTI_INSERT_QSIZE)
						break;
				}
				nstored=i;
				rc=UdmSQLQuery(db,NULL,qb);
				free(qb);
				if(rc!=UDM_OK) goto unlock_StoreWordsSingle;
			}
		}
	}else{
		for(i=0;i<Doc->Words.nwords;i++){
			if(!Doc->Words.Word[i].coord)continue;
				
			if(db->DBMode==UDM_DBMODE_SINGLE){
				sprintf(qbuf,"INSERT INTO dict (url_id,word,intag) VALUES(%d,'%s',%d)",url_id,Doc->Words.Word[i].word,Doc->Words.Word[i].coord);
			}else{
				sprintf(qbuf,"INSERT INTO ndict (url_id,word_id,intag) VALUES(%d,%d,%d)",url_id,UdmStrCRC32(Doc->Words.Word[i].word),Doc->Words.Word[i].coord);
			}
			if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
				goto unlock_StoreWordsSingle;
		}
	}

unlock_StoreWordsSingle:
	/* Commit */
	if(1){
		switch(db->DBType){
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"END WORK");
				break;
			case UDM_DB_MSSQL:
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				break;
			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
			case UDM_DB_SAPDB:
				db->commit_fl = 0;
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				break;
			case UDM_DB_IBASE:
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				db->commit_fl = 1;
				break;
			default:
#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB || HAVE_DB2)
				rc=UdmSQLQuery(db,NULL,"UDM_COMMIT");
#endif
				db->commit_fl = 0;
				break;
		}
		if(rc!=UDM_OK)return rc;
	}
	return(UDM_OK);
}



static int UdmResStoreBlob(UDM_AGENT *Indexer,UDM_RESULT *Res,UDM_DB *db){
	size_t			tabnum,docnum,wnum;
	int			ts=(int)time(NULL);
	char			*blob;
	char			*qb,*e;
	int			*dellog;
	size_t			nbytes;
	UDM_URLCRDLISTLIST	CLst;
	int			rc=UDM_OK;
	
	if(!Res->num_rows)return UDM_OK;
	if(db->DBMode!=UDM_DBMODE_BLOB)return UDM_OK;
	
	bzero((void*)&CLst, sizeof(CLst));
	/* Collect words from all documents */
	UdmResCollectWords(&Indexer->Conf->Indexed,&CLst);
	
	
	/* First of all insert delay log */
	
	switch(db->DBType){
	case UDM_DB_MYSQL:
		nbytes=Res->num_rows*sizeof(int);
		dellog=(int*)malloc(nbytes);
		for(docnum=0;docnum<Res->num_rows;docnum++){
			dellog[docnum]=UdmVarListFindInt(&Res->Doc[docnum].Sections,"ID",0);
		}
		blob=(char*)malloc(1+2*nbytes);
		qb=(char*)malloc(256+2*nbytes);
		UdmDBEscStr(db->DBType,blob,(char*)dellog,nbytes);
		UDM_FREE(dellog);
		sprintf(qb,"INSERT INTO dellog (tstamp,ndocs,docs) VALUES (%d,%d,'%s')",ts,Res->num_rows,blob);
		rc=UdmSQLQuery(db,NULL,qb);
		UDM_FREE(blob);
		UDM_FREE(qb);
		
	case UDM_DB_PGSQL:
	case UDM_DB_IBASE:
		nbytes=Res->num_rows*12+1; /* 0x80000000=-2147483648=11 chars */
		e=blob=(char*)malloc(nbytes);
		qb=(char*)malloc(nbytes+256);
		for(docnum=0;docnum<Res->num_rows;docnum++){
			if(e>blob)*e++=',';
			e+=sprintf(e,"%d",UdmVarListFindInt(&Res->Doc[docnum].Sections,"ID",0));
		}
		sprintf(qb,"INSERT INTO dellog (tstamp,ndocs,docs) VALUES (%d,%d,'{%s}')",ts,Res->num_rows,blob);
		rc=UdmSQLQuery(db,NULL,qb);
		free(blob);
		free(qb);
		break;
	default:
		sprintf(db->errstr,"Blob mode not supported by this DBType");
		db->errcode=1;
		rc=UDM_ERROR;
	}
	
	if(rc!=UDM_OK)goto ret;
	
	/* Insert words */
	for(tabnum=0;tabnum<CLst.nlists;tabnum++){
		UDM_URLCRDLIST	*Lst=&CLst.List[tabnum];
		int		table,crc32;
		
		if(!Lst->ncoords)continue;
		
		crc32=UdmStrCRC32(Lst->word);
		table=(unsigned)crc32 % db->numtables;
		
		switch(db->DBType){
		case UDM_DB_MYSQL:
			nbytes=sizeof(UDM_URL_CRD)*Lst->ncoords;
			blob=(char*)malloc(1+2*nbytes);
			qb=(char*)malloc(256+2*nbytes);
			UdmDBEscStr(db->DBType,blob,(char*)Lst->Coords,nbytes);
			sprintf(qb,"INSERT INTO bdict%02X (word,tstamp,ncoords,coords) VALUES ('%s',%d,%d,'%s')",table,Lst->word,ts,Lst->ncoords,blob);
			rc=UdmSQLQuery(db,NULL,qb);
			UDM_FREE(blob);
			UDM_FREE(qb);
			break;
		case UDM_DB_IBASE:
		case UDM_DB_PGSQL:
			nbytes=Lst->ncoords*sizeof(UDM_URL_CRD)*12+1;
			e=blob=(char*)malloc(nbytes);
			qb=(char*)malloc(256+nbytes);
			for(wnum=0;wnum<Lst->ncoords;wnum++){
				if(e!=blob)*e++=',';
				e+=sprintf(e,"%d,%d",Lst->Coords[wnum].url_id,Lst->Coords[wnum].coord);
			}
			if(Lst->ncoords<1024){
				sprintf(qb,"INSERT INTO bdict%02X (word,tstamp,ncoords,coords) VALUES ('%s',%d,%d,'{%s}')",table,Lst->word,ts,Lst->ncoords,blob);
				rc=UdmSQLQuery(db,NULL,qb);
			}else{
				printf("FIXME SKIP: too long: '%s' %d coords\n",Lst->word,Lst->ncoords);
			}
			free(blob);
			free(qb);
			break;
		}
		if(rc!=UDM_OK)goto ret;
	}
ret:
	UdmURLCRDListListFree(&CLst);
	return rc;
}


static int UdmStoreWords(UDM_AGENT * Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	int	res;
	
	switch(db->DBMode){
		case UDM_DBMODE_BLOB:
			res=UDM_OK;
			break;
		case UDM_DBMODE_CACHE:
			res=UdmStoreWordsCache(Indexer,Doc,db);
			break;
		case UDM_DBMODE_MULTI:
		case UDM_DBMODE_MULTI_CRC:
			res=StoreWordsMulti(Indexer,Doc,db);
			break;
		case UDM_DBMODE_SINGLE:
		case UDM_DBMODE_SINGLE_CRC:
		default:
			res=StoreWordsSingle(Indexer,Doc,db);
			break;
	}
	return(res);
}


static int UdmDeleteAllFromDict(UDM_AGENT *Indexer,UDM_DB *db){
	char	qbuf[512];
	size_t	i,last=0;
	int	rc=UDM_OK;
	
	switch(db->DBMode){
	case UDM_DBMODE_MULTI:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				if (db->DBSQL_TRUNCATE)
					sprintf(qbuf,"TRUNCATE TABLE dict%d",DICTNUM(i));
				else
					sprintf(qbuf,"DELETE FROM dict%d",DICTNUM(i));

				if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
					return rc;
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_MULTI_CRC:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				if (db->DBSQL_TRUNCATE)
					sprintf(qbuf,"TRUNCATE TABLE ndict%d",DICTNUM(i));
				else
					sprintf(qbuf,"DELETE FROM ndict%d",DICTNUM(i));
	
				if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
					return(UDM_ERROR);
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_SINGLE_CRC:
		if (db->DBSQL_TRUNCATE)
			rc=UdmSQLQuery(db,NULL,"TRUNCATE TABLE ndict");
		else
			rc=UdmSQLQuery(db,NULL,"DELETE FROM ndict");
		break;
	case UDM_DBMODE_BLOB:
		for(i=0;i<db->numtables;i++){
			if (db->DBSQL_TRUNCATE)
				sprintf(qbuf,"TRUNCATE TABLE bdict%02X",i);
			else
				sprintf(qbuf,"DELETE FROM bdict%02X",i);
			if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
				return rc;
		}
		rc=UdmSQLQuery(db,NULL,"DELETE FROM dellog");
		break;
	default:
		if (db->DBSQL_TRUNCATE)
			rc=UdmSQLQuery(db,NULL,"TRUNCATE TABLE dict");
		else
			rc=UdmSQLQuery(db,NULL,"DELETE FROM dict");
		break;
	}
	return rc;
}


/***************** CrossWords *******************************/

static int UdmDeleteAllFromCrossDict(UDM_AGENT * Indexer,UDM_DB *db){
	char	qbuf[1024];
	char	table[64]="ncrossdict";
	int	crcmode=1;
	
	if((db->DBMode==UDM_DBMODE_SINGLE)||(db->DBMode==UDM_DBMODE_MULTI)||(db->DBMode==UDM_DBMODE_CACHE)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	sprintf(qbuf,"DELETE FROM %s",table);
	return UdmSQLQuery(db,NULL,qbuf);
}


static int UdmDeleteCrossWordFromURL(UDM_AGENT * Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	char	qbuf[1024];
	char	table[64]="ncrossdict";
	int	crcmode=1;
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int	referrer_id=UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0);
	int	rc=UDM_OK;
	
	if((db->DBMode==UDM_DBMODE_SINGLE)||(db->DBMode==UDM_DBMODE_MULTI)||(db->DBMode==UDM_DBMODE_CACHE)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	if(url_id){
		sprintf(qbuf,"DELETE FROM %s WHERE url_id=%d",table,url_id);
		if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
			return rc;
	}
	if(referrer_id){
		sprintf(qbuf,"DELETE FROM %s WHERE ref_id=%d",table,referrer_id);
		rc=UdmSQLQuery(db,NULL,qbuf);
	}
	return rc;
}


static int UdmStoreCrossWords(UDM_AGENT * Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	UDM_DOCUMENT	U;
	size_t		i;
	char		qbuf[1024];
	char		table[64]="ncrossdict";
	int		crcmode=1;
	const char	*lasturl="scrap";
	int		referrer=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int		childid=0;
	int		rc=UDM_OK;
	UDM_HREF        Href;
	UDM_URL         docURL;
	
	UdmDocInit(&U);
	bzero((void*)&Href, sizeof(Href));
	UdmVarListReplaceInt(&Doc->Sections,"Referrer-ID",referrer);
	if(UDM_OK!=(rc=UdmDeleteCrossWordFromURL(Indexer,&U,db))){
		UdmDocFree(&U);
		return rc;
	}
	
	if(Doc->CrossWords.ncrosswords==0) {
		UdmDocFree(&U);
		return rc;
	}
	
	if((db->DBMode==UDM_DBMODE_SINGLE)||(db->DBMode==UDM_DBMODE_MULTI)||(db->DBMode==UDM_DBMODE_CACHE)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	

	UdmURLParse(&docURL, UdmVarListFindStr(&Doc->Sections, "URL", ""));
	for(i=0;i<Doc->CrossWords.ncrosswords;i++){
		if(!Doc->CrossWords.CrossWord[i].weight)continue;
		if(strcmp(lasturl,Doc->CrossWords.CrossWord[i].url)){
		        Href.url = strdup(Doc->CrossWords.CrossWord[i].url);
			UdmConvertHref(Indexer, &docURL, &Href);
			UdmVarListReplaceStr(&U.Sections, "URL", Href.url);
			if(UDM_OK!=(rc=UdmFindURL(Indexer,&U,db))){
				UdmDocFree(&U);
				return rc;
			}
			childid=UdmVarListFindInt(&U.Sections,"ID",0);
			lasturl=Doc->CrossWords.CrossWord[i].url;
			UDM_FREE(Href.url);
		}
		Doc->CrossWords.CrossWord[i].referree_id=childid;
	}
	
	/* Begin transacttion/lock */
	if(1){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				sprintf(qbuf,"LOCK TABLES %s WRITE",table);
				rc=UdmSQLQuery(db,NULL,qbuf);
				break;
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"BEGIN WORK");
				break;
			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
			case UDM_DB_SAPDB:
				db->commit_fl = 1;
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				break;
			default:
				db->commit_fl = 1;
				break;
		}
		if(rc!=UDM_OK){
			UdmDocFree(&U);
			return rc;
		}
	}
	
	/* Insert new words */
	for(i=0;i<Doc->CrossWords.ncrosswords;i++){
		if(Doc->CrossWords.CrossWord[i].weight && Doc->CrossWords.CrossWord[i].referree_id){
			int weight=UDM_WRDCOORD(Doc->CrossWords.CrossWord[i].pos,Doc->CrossWords.CrossWord[i].weight);
			if(crcmode){
				sprintf(qbuf,"INSERT INTO %s (ref_id,url_id,word_id,intag) VALUES(%d,%d,%d,%d)",table,referrer,Doc->CrossWords.CrossWord[i].referree_id,UdmStrCRC32(Doc->CrossWords.CrossWord[i].word),weight);
			}else{
				sprintf(qbuf,"INSERT INTO %s (ref_id,url_id,word,intag) VALUES(%d,%d,'%s',%d)",table,referrer,Doc->CrossWords.CrossWord[i].referree_id,Doc->CrossWords.CrossWord[i].word,weight);
			}
			if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf))){
				UdmDocFree(&U);
				goto unlock_UdmStoreCrossWords;
			}
		}
	}

unlock_UdmStoreCrossWords:
	/* COMMIT/UNLOCK */
	if(1){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				rc=UdmSQLQuery(db,NULL,"UNLOCK TABLES");
				break;
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"END WORK");
				break;
			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
			case UDM_DB_SAPDB:
				db->commit_fl = 0;
				rc=UdmSQLQuery(db,NULL,"COMMIT");
				break;
			default:
#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB || HAVE_DB2)
				rc=UdmSQLQuery(db,NULL,"UDM_COMMIT");
#endif
				db->commit_fl = 0;
				break;
		}
	}
	UdmDocFree(&U);
	return rc;
}




/************************ URLs ***********************************/


static int UdmAddURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc,UDM_DB *db){
	char		e_url[UDM_URLSIZE*4]="";
	int		next_url_id=0;
	char		qbuf[UDM_URLSIZE*4]="AddURL";
	UDM_SQLRES	SQLRes;
	const char	*url=UdmVarListFindStr(&Doc->Sections,"URL","");
	int		url_seed;
	int		use_crc32_url_id=!strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"UseCRC32URLId","no"),"yes");
	int		rc=UDM_OK;
	
	url_seed = UdmStrCRC32(url) & 0xFF;
	
	/* Escape URL string */
	UdmDBEscStr(db->DBType,e_url,url,strlen(url));
	
	if(use_crc32_url_id){
		/* Auto generation of rec_id */
		/* using CRC32 algorythm     */
		int rec_id=UdmStrCRC32(url);
		
		snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (rec_id,url,referrer,hops,crc32,next_index_time,status,seed,bad_since_time,site_id,server_id) VALUES (%d,'%s',%d,%d,0,%d,0,%d,%d,%d,%d)",
			rec_id,
			e_url,
			UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
			UdmVarListFindInt(&Doc->Sections,"Hops",0),
			(int)time(NULL),
			 url_seed, (int)time(NULL),
			 UdmVarListFindInt(&Doc->Sections,"Site_id",0),
			 UdmVarListFindInt(&Doc->Sections,"Server_id",0)
			 );
	}else{
		/* Use dabatase generated rec_id */
		/* It depends on used DBType     */
		switch(db->DBType){
		case UDM_DB_MSQL:
			/* miniSQL has _seq as autoincrement value */
			if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes,"SELECT _seq FROM url")))
				return rc;
			next_url_id=atoi(UdmSQLValue(&SQLRes,0,0));
			UdmSQLFree(&SQLRes);
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,site_id,server_id) VALUES ('%s',%d,%d,%d,0,%d,0,%d,%d,%d,%d)",
				e_url,
				UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
				UdmVarListFindInt(&Doc->Sections,"Hops",0),
				next_url_id,
				(int)time(NULL),
				 url_seed, (int)time(NULL),
				 UdmVarListFindInt(&Doc->Sections,"Site_id",0),
				 UdmVarListFindInt(&Doc->Sections,"Server_id",0)
				 );
			break;

		case UDM_DB_SOLID:
		case UDM_DB_ORACLE7:
		case UDM_DB_ORACLE8:
		case UDM_DB_SAPDB:
			/* FIXME: Dirty hack for stupid too smart databases */
			if (strlen(e_url)>UDM_URLSIZE)e_url[UDM_URLSIZE]=0;
			/* Use sequence next_url_id.nextval */
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,site_id,server_id) VALUES ('%s',%d,%d,next_url_id.nextval,0,%d,0,%d,%d,%d,%d)",
				e_url,
				UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
				UdmVarListFindInt(&Doc->Sections,"Hops",0),
				(int)time(NULL),
				 url_seed, (int)time(NULL),
				 UdmVarListFindInt(&Doc->Sections,"Site_id",0),
				 UdmVarListFindInt(&Doc->Sections,"Server_id",0)
				 );
			break;
		case UDM_DB_IBASE:
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,rec_id,crc32,next_index_time,status,seed,bad_since_time,site_id,server_id) VALUES ('%s',%d,%d,GEN_ID(rec_id_GEN,1),0,%d,0,%d,%d,%d,%d)",
				e_url,
				UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
				UdmVarListFindInt(&Doc->Sections,"Hops",0),
				(int)time(NULL),
				 url_seed, (int)time(NULL),
				 UdmVarListFindInt(&Doc->Sections,"Site_id",0),
				 UdmVarListFindInt(&Doc->Sections,"Server_id",0)
				 );
			break;
		case UDM_DB_MYSQL:
			/* MySQL generates itself */
		default:	
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,crc32,next_index_time,status,seed,bad_since_time,site_id,server_id) VALUES ('%s',%d,%d,0,%d,0,%d,%d,%d,%d)",
				e_url,
				UdmVarListFindInt(&Doc->Sections,"Referrer-ID",0),
				UdmVarListFindInt(&Doc->Sections,"Hops",0),
				(int)time(NULL),
				 url_seed, (int)time(NULL),
				 UdmVarListFindInt(&Doc->Sections,"Site_id",0),
				 UdmVarListFindInt(&Doc->Sections,"Server_id",0)
				 );
		}
	}

	/* Exec INSERT now */
	if(UDM_OK!=(rc=UdmSQLQuery(db, NULL, qbuf)))
		return rc;

	snprintf(qbuf, sizeof(qbuf) - 1, "SELECT rec_id FROM url WHERE url='%s'", e_url);
	if(UDM_OK != (rc = UdmSQLQuery(db, &SQLRes, qbuf)))
		return rc;

	if (UdmSQLNumRows(&SQLRes)) {
		snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO links (ot,k) VALUES (%d, %s)",
			UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0),
			UdmSQLValue(&SQLRes, 0, 0));
	} else {
		UdmLog(Indexer, UDM_LOG_ERROR, "URL not found: %s", e_url);
		return UDM_ERROR;
	}

	UdmSQLFree(&SQLRes);
	return UdmSQLQuery(db, NULL, qbuf);
}

static int UdmDeleteURL(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db);

static int UdmDeleteBadHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_DB *db) {
	UDM_DOCUMENT	rDoc;
	UDM_SQLRES	SQLRes;
	char		q[128];
	size_t		i;
	size_t		nrows;
	int		rc=UDM_OK;
	int		hold_period=UdmVarListFindInt(&Doc->Sections,"HoldBadHrefs",0);
	int		url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	
	sprintf(q, "SELECT rec_id FROM url WHERE status >=300 AND status<>304 AND (referrer=%d OR referrer=0) AND bad_since_time<%d",
		url_id, (int)time(NULL) - hold_period);
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes,q)))return rc;
	
	nrows = UdmSQLNumRows(&SQLRes);
	
	UdmDocInit(&rDoc);
	for(i = 0; i < nrows ; i++) {
		UdmVarListReplaceStr(&rDoc.Sections,"ID", UdmSQLValue(&SQLRes,i,0));
		if(UDM_OK!=(rc=UdmDeleteURL(Indexer, &rDoc, db)))
			break;
	}
	UdmDocFree(&rDoc);
	UdmSQLFree(&SQLRes);
	return rc;
}

static int UdmDeleteURL(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db){
	char	qbuf[128];
	int	rc;
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int	use_crosswords=!strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"CrossWords","no"),"yes");
	
	if(use_crosswords&&db->DBMode!=UDM_DBMODE_CACHE)
		if(UDM_OK!=(rc=UdmDeleteCrossWordFromURL(Indexer,Doc,db)))return(rc);
	
	if(UDM_OK!=(rc=UdmDeleteWordFromURL(Indexer,Doc,db)))return(rc);
	
	sprintf(qbuf,"DELETE FROM url WHERE rec_id=%d",url_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))return rc;
	
	sprintf(qbuf,"DELETE FROM urlinfo WHERE url_id=%d",url_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))return rc;
	
	sprintf(qbuf,"DELETE FROM links WHERE ot=%d", url_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))return rc;
	
	sprintf(qbuf,"DELETE FROM links WHERE k=%d", url_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))return rc;
	
	UdmStoreDeleteDoc(Indexer, Doc);

	/* remove all old broken hrefs from this document to avoid broken link collecting */
	if (UDM_OK != (rc=UdmDeleteBadHrefs(Indexer,Doc, db))) return rc;

	sprintf(qbuf,"UPDATE url SET referrer = 0 WHERE referrer=%d", url_id);
	return UdmSQLQuery(db,NULL,qbuf);
}

static int UdmDeleteLinks(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc, UDM_DB *db){
	char	qbuf[128];
	int	url_id = UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);

	sprintf(qbuf,"DELETE FROM links WHERE ot=%d", url_id);
	return UdmSQLQuery(db, NULL, qbuf);
}

static int UdmMarkForReindex(UDM_AGENT *Indexer,UDM_DB *db){
	char		qbuf[1024];
	char            ubuf[4096];
	const char	*where=BuildWhere(Indexer->Conf,db);
	UDM_SQLRES 	SQLRes;
	size_t          i, j;
	int             rc;
	
	if (db->DBSQL_SUBSELECT) {
	  snprintf(qbuf,sizeof(qbuf),"UPDATE url SET next_index_time=%d WHERE rec_id IN (SELECT url.rec_id FROM url%s %s %s)",
		   (int)time(NULL), db->from, (where[0]) ? "WHERE" : "", where);
	  return UdmSQLQuery(db,NULL,qbuf);
	}
	
	snprintf(qbuf, sizeof(qbuf), "SELECT url.rec_id FROM url%s %s %s", db->from, (where[0]) ? "WHERE" : "", where);
	if(UDM_OK != (rc = UdmSQLQuery(db, &SQLRes, qbuf))) return rc;
	
	if (db->DBSQL_IN) {
	  for (i = 0; i < UdmSQLNumRows(&SQLRes); i += 512) {
	    sprintf(ubuf, "UPDATE url SET next_index_time=%d WHERE rec_id IN (", (int)time(NULL));
	    for (j = 0; (j < 512) && (i + j < UdmSQLNumRows(&SQLRes)); j++) {
	      sprintf(UDM_STREND(ubuf), "%s%s", (j) ? "," : "", UdmSQLValue(&SQLRes, i + j, 0));
	    }
	    sprintf(UDM_STREND(ubuf), ")");
	    if(UDM_OK != (rc = UdmSQLQuery(db, NULL, ubuf))) {
	      UdmSQLFree(&SQLRes);
	      return rc;
	    }
	  }
	} else {
	  for (i = 0; i < UdmSQLNumRows(&SQLRes); i++) {
	    sprintf(ubuf, "UPDATE url SET next_index_time=%d WHERE rec_id=%s", (int)time(NULL),  UdmSQLValue(&SQLRes, i, 0));
	    if(UDM_OK != (rc = UdmSQLQuery(db, NULL, ubuf))) {
	      UdmSQLFree(&SQLRes);
	      return rc;
	    }
	  }
	}
	UdmSQLFree(&SQLRes);
	return UDM_OK;
}

static int UdmUpdateUrl(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	char	qbuf[256]="";
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int	status=UdmVarListFindInt(&Doc->Sections,"Status",0);
	int	prevStatus = UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0);
	int	next_index_time=UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Next-Index-Time",""));
	int	res;
	
#ifdef HAVE_ORACLE8
	if(db->DBDriver==UDM_DB_ORACLE8){
	  if ( (prevStatus != status) && (status > 300) && (status != 304) ) {
		sprintf(qbuf,"UPDATE url SET status=:1,next_index_time=:2,bad_since_time=:4, site_id=%d, server_id=%d WHERE rec_id=:3",
			UdmVarListFindInt(&Doc->Sections, "Site_id",0), UdmVarListFindInt(&Doc->Sections, "Server_id",0));
		param_init(db, 1, 2, 3, 4);
		param_add(db, status, next_index_time, url_id, time(NULL));
	  } else {
		sprintf(qbuf,"UPDATE url SET status=:1,next_index_time=:2, site_id=%d WHERE rec_id=:3",
			UdmVarListFindInt(&Doc->Sections, "Site_id",0));
		param_init(db, 1, 2, 3, 0);
		param_add(db, status, next_index_time, url_id, 0);
	  }
	}
#endif
	if(!qbuf[0]) {
	  if ( (prevStatus != status) && (status > 300) && (status != 304) )
		sprintf(qbuf, "UPDATE url SET status=%d,next_index_time=%d,bad_since_time=%d,site_id=%d,server_id=%d WHERE rec_id=%d",
			status, next_index_time, (int)time(NULL),UdmVarListFindInt(&Doc->Sections, "Site_id", 0),
			UdmVarListFindInt(&Doc->Sections, "Server_id",0), url_id);
	  else
		sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=%d, site_id=%d,server_id=%d WHERE rec_id=%d",status,next_index_time,
			UdmVarListFindInt(&Doc->Sections, "Site_id", 0), UdmVarListFindInt(&Doc->Sections, "Server_id",0), url_id);
	}
	
	if(UDM_OK!=(res=UdmSQLQuery(db,NULL,qbuf)))return res;
	
	/* remove all old broken hrefs from this document to avoid broken link collecting */
	return UdmDeleteBadHrefs(Indexer,Doc,db);
}

static int UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db)
{
	char	qbuf[1024];
	int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int	parent_id=UdmVarListFindInt(&Doc->Sections,"Parent-ID",0);
	
	snprintf(qbuf,sizeof(qbuf),"insert into links (ot, k) values(%d,%d)",parent_id, url_id);
	return UdmSQLQuery(db,NULL,qbuf);
}



static int  UdmUpdateClone(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db){
	char	*qbuf;
	int	rc;
	const char	*charset;
	UDM_VAR		*var;
	int		status, prevStatus, url_id;
	size_t		i, len = 0;
	char		qsmall[64];
	
	status = UdmVarListFindInt(&Doc->Sections, "Status", 0);
	prevStatus = UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0);
	url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
	
#if 0
	{
		int a;
		char *l=UdmVarListFindStr(&Doc->Sections,"Last-Modified","");
		a=UdmHttpDate2Time_t(l);
		fprintf(stderr,"DDD='%s'\n%d\n",UdmVarListFindStr(&Doc->Sections,"Last-Modified","<NULL>"),a);
	}
#endif
	
	if((var=UdmVarListFind(&Doc->Sections,"Content-Language"))){
	        if (var->val == NULL) {
			var->val = strdup(UdmVarListFindStr(&Doc->Sections, "DefaultLang", "en"));
		}
		len=strlen(var->val);
		for(i = 0; i < len; i++) {
			var->val[i] = tolower(var->val[i]);
		}
	}
	
	charset = UdmVarListFindStr(&Doc->Sections, "Charset", 
				    UdmVarListFindStr(&Doc->Sections, "RemoteCharset", "iso-8859-1"));
	charset = UdmCharsetCanonicalName(charset);
	UdmVarListReplaceStr(&Doc->Sections, "Charset", charset);
	
	if ( (prevStatus != status) && (status > 300) && (status != 304) )
	  snprintf(qsmall, 64, ", bad_since_time=%d", (int)time(NULL));
	else qsmall[0] = '\0';
	
	qbuf=(char*)malloc(1024);
	
	
snprintf(qbuf, 1023, "\
UPDATE url SET \
status=%d,last_mod_time=%li,\
next_index_time=%li,\
docsize=%d,\
crc32=%d%s, site_id=%d, server_id=%d \
WHERE rec_id=%d",
	UdmVarListFindInt(&Doc->Sections,"Status",0),
	UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Last-Modified","")),
	UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Next-Index-Time","")),
	UdmVarListFindInt(&Doc->Sections,"Content-Length",0),
	UdmVarListFindInt(&Doc->Sections,"crc32",0),
	qsmall,
	UdmVarListFindInt(&Doc->Sections,"Site_id",0),
	UdmVarListFindInt(&Doc->Sections, "Server_id",0),
	url_id);
	
	rc=UdmSQLQuery(db,NULL,qbuf);
	free(qbuf);
	return rc;
}


static int UdmLongUpdateURL(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	int		rc=UDM_OK;
	size_t		i;
	char		*qbuf;
	char		*arg;
	char		qsmall[128];
	size_t		len=0;
	int		url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	
	if(UDM_OK != (rc = UdmUpdateClone(Indexer, Doc, db)))
		return rc;
	
	/* remove all old broken hrefs from this document to avoid broken link collecting */
	if(UDM_OK!=(rc=UdmDeleteBadHrefs(Indexer,Doc,db)))
		return rc;
	
	sprintf(qsmall,"DELETE FROM urlinfo WHERE url_id=%d",url_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qsmall)))return rc;

/* No need delete from links here, it has been done before */
	
	if(!Doc->Sections.nvars) return UDM_OK;
	
	len=0;
	for(i=0;i<Doc->Sections.nvars;i++){
	        size_t l = Doc->Sections.Var[i].curlen + (Doc->Sections.Var[i].name ? strlen(Doc->Sections.Var[i].name) : 0);
		if(len < l)
			len = l;
	}
	if(!len)return UDM_OK;
	
	qbuf=(char*)malloc(2 * len + 128);
	arg=(char*)malloc(2 * len + 128);
	
	for(i=0;i<Doc->Sections.nvars;i++){
		UDM_VAR *Sec=&Doc->Sections.Var[i];
		if(Sec->curlen && Sec->val && Sec->name && Sec->section){
			arg=UdmDBEscStr(db->DBType,arg,Sec->val,strlen(Sec->val));
			sprintf(qbuf,"INSERT INTO urlinfo (url_id,sname,sval) VALUES (%d,'%s','%s')",
				url_id,Sec->name,arg);
			if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))break;
		}
	}
	free(qbuf);
	free(arg);
	return rc;
}


static int UdmDeleteAllFromUrl(UDM_AGENT *Indexer,UDM_DB *db){
	int	rc;
	
	if(db->DBSQL_TRUNCATE)
		rc=UdmSQLQuery(db,NULL,"TRUNCATE TABLE url");
	else
		rc=UdmSQLQuery(db,NULL,"DELETE FROM url");
	
	if(rc!=UDM_OK)return rc;
	
	if(db->DBSQL_TRUNCATE)
		rc = UdmSQLQuery(db, NULL, "TRUNCATE TABLE links");
	else
		rc = UdmSQLQuery(db, NULL, "DELETE FROM links");
	
	if(rc != UDM_OK) return rc;
	
	if(db->DBSQL_TRUNCATE)
		rc=UdmSQLQuery(db,NULL,"TRUNCATE TABLE urlinfo");
		
	else
		rc=UdmSQLQuery(db,NULL,"DELETE FROM urlinfo");
	
	return rc;
}



/************************ Clones stuff ***************************/
static int UdmFindOrigin(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc,UDM_DB *db){
	size_t		i=0;
	char		qbuf[256]="";
	UDM_SQLRES	SQLRes;
	int		origin_id=0;
	int		crc32=UdmVarListFindInt(&Doc->Sections,"crc32",0);
	int		rc;
	
	if (crc32==0)return UDM_OK;
	
	if (db->DBSQL_IN)
		sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=%d AND status IN (200,304,206)",crc32);
	else
		sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206)",crc32);
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes,qbuf)))
		return rc;
	
	for(i=0;i<UdmSQLNumRows(&SQLRes);i++){
		const char *o;
		if((o=UdmSQLValue(&SQLRes,i,0)))
			if((!origin_id)||(origin_id>atoi(o)))
				origin_id=atoi(o);
	}
	UdmSQLFree(&SQLRes);
	UdmVarListReplaceInt(&Doc->Sections,"Origin-ID",origin_id);
	return(UDM_OK);
}



/************** Get Target to be indexed ***********************/

static int UdmTargets(UDM_AGENT *Indexer,UDM_RESULT *Res,UDM_DB *db){
	char		sortstr[128] = "";
	char		updstr[64]="";
	char		lmtstr[64]="";
	size_t		i = 0, j;
	size_t		url_num = UdmVarListFindInt(&Indexer->Conf->Vars, "URLSelectCacheSize", URL_SELECT_CACHE);
	UDM_SQLRES 	SQLRes;
	char		smallbuf[128];
	int		rc=UDM_OK;
	const char	*where=BuildWhere(Indexer->Conf,db);
	char		*qbuf=NULL;

	UdmResultFree(&Indexer->Conf->Targets);
	
	if(1){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				rc=UdmSQLQuery(db,NULL,"LOCK TABLES url WRITE");
				break;
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"BEGIN WORK");
				rc=UdmSQLQuery(db,NULL,"LOCK url");
				break;
			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
				sprintf(updstr, " FOR UPDATE ");
#if HAVE_ORACLE8
				if(db->DBDriver==UDM_DB_ORACLE8){
					sprintf(lmtstr, " AND ROWNUM <=%d",url_num); 
				}
#endif
				if(!lmtstr[0])
					sprintf(lmtstr, " AND ROWNUM <=%d", url_num); 
				break;
			case UDM_DB_SAPDB:
				sprintf(updstr, " WITH LOCK ");
				strcpy(lmtstr, "");
				break;
			default:
				break;
		}
		if(rc!=UDM_OK)return rc;
	}
	
		
	if ( (qbuf = (char*)malloc(512 + 4 * UDM_URLSIZE)) == NULL) {
		  UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
		  rc = UDM_ERROR;
		  goto unlock;
	}
	qbuf[0]='\0';

	if ( (Indexer->flags & (UDM_FLAG_SORT_HOPS | UDM_FLAG_SORT_EXPIRED)) || !(Indexer->flags & UDM_FLAG_DONTSORT_SEED)  ) {
		
	  sprintf(sortstr, " ORDER BY %s%s%s", 
		  (Indexer->flags & UDM_FLAG_SORT_HOPS) ? "hops" : "",
		  (Indexer->flags & UDM_FLAG_DONTSORT_SEED) ? "" : ((Indexer->flags & UDM_FLAG_SORT_HOPS) ? ",seed" : "seed"),
		  (Indexer->flags & UDM_FLAG_SORT_EXPIRED) ? 
		  ( ((Indexer->flags & UDM_FLAG_SORT_HOPS) || !(Indexer->flags & UDM_FLAG_DONTSORT_SEED)  ) ? 
		    ",next_index_time" : "next_index_time") : "");
	} else {
	  sortstr[0] = '\0';
	}
		
			
	if(db->DBSQL_LIMIT){
	  sprintf(qbuf,"SELECT url.url,url.rec_id,docsize,status,hops,crc32,last_mod_time,seed FROM url%s WHERE next_index_time<=%d %s %s %s LIMIT %d",
		  db->from, (int)time(NULL), where[0] ? "AND" : "", where, sortstr, url_num);
	}else{
	  db->res_limit=url_num;
/*
#if HAVE_ORACLE8
					if(db->DBDriver==UDM_DB_ORACLE8){
						param_init(db, 1, 2, 0, 0);
						param_add(db, (int)time(NULL), url_num, 0, 0);
						sprintf(qbuf,"SELECT url,rec_id,docsize,status,hops,crc32,last_mod_time FROM url WHERE next_index_time<=:1 %s %s%s%s%s",
							where[0]?"AND":"",where,lmtstr,sortstr,updstr);
					}
#endif
*/
	  if(!qbuf[0])
	    sprintf(qbuf,"SELECT url.url,url.rec_id,docsize,status,hops,crc32,last_mod_time FROM url%s WHERE next_index_time<=%d %s %s%s%s%s",
		    db->from, (int)time(NULL), where[0] ? "AND" : "", where, lmtstr, sortstr, updstr);
	}
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLRes, qbuf))) goto unlock;
	
	Indexer->Conf->Targets.num_rows=UdmSQLNumRows(&SQLRes);
	
	if(!Indexer->Conf->Targets.num_rows){
		UdmSQLFree(&SQLRes);
		goto unlock;
	}
	
	Indexer->Conf->Targets.Doc = (UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Indexer->Conf->Targets.num_rows));
	
	for(i=0;i<Indexer->Conf->Targets.num_rows;i++){
		char		buf[64]="";
		time_t		last_mod_time;
		UDM_DOCUMENT	*Doc=&Indexer->Conf->Targets.Doc[i];
		
		UdmVarListAddStr(&Doc->Sections,"URL",UdmSQLValue(&SQLRes,i,0));
		UdmVarListAddInt(&Doc->Sections,"ID",atoi(UdmSQLValue(&SQLRes,i,1)));
		UdmVarListAddInt(&Doc->Sections,"Status",atoi(UdmSQLValue(&SQLRes,i,3)));
		UdmVarListAddInt(&Doc->Sections,"PrevStatus",atoi(UdmSQLValue(&SQLRes,i,3)));
		UdmVarListAddInt(&Doc->Sections,"Hops",atoi(UdmSQLValue(&SQLRes,i,4)));
		UdmVarListAddInt(&Doc->Sections,"crc32",atoi(UdmSQLValue(&SQLRes,i,5)));
		last_mod_time = (time_t) atol(UdmSQLValue(&SQLRes,i,6));
		UdmTime_t2HttpStr(last_mod_time, buf);
		if (strlen(buf) > 0) {
		  UdmVarListReplaceStr(&Doc->Sections,"Last-Modified",buf);
		}
	}
	UdmSQLFree(&SQLRes);
	
	
	if (db->DBSQL_IN) {
		char	*urlin=NULL;
		
		if ( (qbuf = (char*)realloc(qbuf, 512 + 4 * UDM_URLSIZE + 35 * URL_SELECT_CACHE)) == NULL) {
			UDM_FREE(qbuf);
			UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
			rc= UDM_ERROR;
			goto unlock;
		}
		
		if ( (urlin = (char*)malloc(35 * URL_SELECT_CACHE)) == NULL) {
			UDM_FREE(qbuf);
			UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory");
			rc = UDM_ERROR;
			goto unlock;
		}
		urlin[0]=0;
		
		for(i = 0; i < Indexer->Conf->Targets.num_rows; i+= URL_SELECT_CACHE) {

		  urlin[0] = 0;

		  for (j = 0; (j < URL_SELECT_CACHE) && (i + j < Indexer->Conf->Targets.num_rows) ; j++) {

			UDM_DOCUMENT	*Doc = &Indexer->Conf->Targets.Doc[i + j];
			int		url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
			
			if(urlin[0])strcat(urlin,",");
			sprintf(urlin+strlen(urlin),"%d",url_id);
		  }
		  sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id in (%s)",(int)(time(NULL)+URL_LOCK_TIME),urlin);
		  rc = UdmSQLQuery(db,NULL,qbuf);
		  if (rc != UDM_OK) goto unlock;
		}
		free(urlin);
	}else{
		for(i=0;i<Indexer->Conf->Targets.num_rows;i++){
			UDM_DOCUMENT	*Doc=&Indexer->Conf->Targets.Doc[i];
			int		url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
			
			sprintf(smallbuf,"UPDATE url SET next_index_time=%d WHERE rec_id=%d",(int)(time(NULL)+URL_LOCK_TIME),url_id);
			if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,smallbuf))) goto unlock;
		}
	}
	
	
unlock:
	if(1){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				rc=UdmSQLQuery(db,NULL,"UNLOCK TABLES");
				break;
			case UDM_DB_PGSQL:
				rc=UdmSQLQuery(db,NULL,"END WORK");
				break;
			default:
				break;
		}
	}
	UDM_FREE(qbuf);
	return rc;
}


/************************************************************/
/* Misc functions                                           */
/************************************************************/

static int UdmGetDocCount(UDM_AGENT * Indexer,UDM_DB *db){
	char		qbuf[200]="";
	UDM_SQLRES	SQLres;
	int		rc;
	
	Indexer->doccount=0;
	sprintf(qbuf,NDOCS_QUERY);
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))return rc;
	
	if(UdmSQLNumRows(&SQLres)){
		const char * s;
		s=UdmSQLValue(&SQLres,0,0);
		if(s)Indexer->doccount=atoi(s);
	}
	UdmSQLFree(&SQLres);
	return(UDM_OK);
}


int UdmStatActionSQL(UDM_AGENT *Indexer,UDM_STATLIST *Stats,UDM_DB *db){
	size_t		i,j;
	char		qbuf[2048];
	UDM_SQLRES	SQLres;
	int		have_group=db->DBSQL_GROUP;
	const char	*where;
	int		rc=UDM_OK;
	
	if(db->DBType==UDM_DB_IBASE)
		have_group=0;
	
	where=BuildWhere(Indexer->Conf,db);
	bzero((void*)Stats, sizeof(Stats[0]));

	if(have_group){
		
		switch(db->DBType){
		
			case UDM_DB_MYSQL:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,sum(next_index_time<=%d),count(*) FROM url%s WHERE url.rec_id<>0 %s %s GROUP BY status",
					(int)time(NULL), db->from, where[0] ? "AND" : "", where);
				break;
		
			case UDM_DB_PGSQL:
			case UDM_DB_MSSQL:
			case UDM_DB_DB2:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,sum(case when next_index_time<=%d then 1 else 0 end),count(*) FROM url%s WHERE url.rec_id<>0 %s %s GROUP BY status",
					(int)time(NULL), db->from, where[0] ? "AND" :"", where);
				break;

			case UDM_DB_ORACLE7:
			case UDM_DB_ORACLE8:
			case UDM_DB_SAPDB:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status, SUM(DECODE(SIGN(%d-next_index_time),-1,0,1,1)), count(*) FROM url%s WHERE url.rec_id<>0 %s %s GROUP BY status",
					 (int)time(NULL), db->from, where[0] ? "AND" : "", where);
				break;
		}

		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))return rc;
		
		if((j=UdmSQLNumRows(&SQLres))){
			for(i=0;i<j;i++){
				UDM_STAT	*S;
				
				Stats->Stat=(UDM_STAT*)realloc(Stats->Stat,(Stats->nstats+1)*sizeof(Stats->Stat[0]));
				S=&Stats->Stat[Stats->nstats];
				S->status=atoi(UdmSQLValue(&SQLres,i,0));
				S->expired=atoi(UdmSQLValue(&SQLres,i,1));
				S->total=atoi(UdmSQLValue(&SQLres,i,2));
				Stats->nstats++;
			}
		}
		UdmSQLFree(&SQLres);
	
	}else{
/*	
	FIXME: learn how to get it from SOLID and IBASE
	(HAVE_IBASE||HAVE_MSQL || HAVE_SOLID || HAVE_VIRT )
*/
		int	n=time(NULL);
		
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,next_index_time FROM url%s WHERE url.rec_id>0 %s %s",
			db->from, where[0] ? "AND" : "", where);

		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))return rc;
		
		for(i=0;i<UdmSQLNumRows(&SQLres);i++){
			for(j=0;j<Stats->nstats;j++){
				if(Stats->Stat[j].status==atoi(UdmSQLValue(&SQLres,i,0))){
					if(atoi(UdmSQLValue(&SQLres,i,1))<=n)
						Stats->Stat[j].expired++;
					Stats->Stat[j].total++;
					break;
				}
			}
			if(j==Stats->nstats){
				Stats->Stat=(UDM_STAT *)realloc(Stats->Stat,sizeof(UDM_STAT)*(Stats->nstats+1));
				Stats->Stat[j].status=atoi(UdmSQLValue(&SQLres,i,0));
				Stats->Stat[j].expired=0;
				if(atoi(UdmSQLValue(&SQLres,i,1))<=n)
					Stats->Stat[j].expired++;
				Stats->Stat[j].total=1;
				Stats->nstats++;
			}
		}
		UdmSQLFree(&SQLres);
	}
	return rc;
}


static int UdmGetReferers(UDM_AGENT *Indexer,UDM_DB *db){
	size_t		i,j;
	char		qbuf[2048];
	UDM_SQLRES	SQLres;
	const char	*where=BuildWhere(Indexer->Conf,db);
	int		rc;
	
	snprintf(qbuf,sizeof(qbuf),"SELECT url.status,url2.url,url.url FROM url,url url2%s WHERE url.referrer=url2.rec_id %s %s",
		db->from, where[0] ? "AND" : "", where);
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))return rc;
	
	j=UdmSQLNumRows(&SQLres);
	for(i=0;i<j;i++){
		if(Indexer->Conf->RefInfo)Indexer->Conf->RefInfo(
			atoi(UdmSQLValue(&SQLres,i,0)),
			UdmSQLValue(&SQLres,i,2),
			UdmSQLValue(&SQLres,i,1)
		);
	}
	UdmSQLFree(&SQLres);
	return rc;
}


int UdmClearDBSQL(UDM_AGENT *Indexer,UDM_DB *db){
	size_t		i,j;
	int		rc;
	char		qbuf[1024*4];
	const char*	where=BuildWhere(Indexer->Conf,db);
/*	int		use_newsext=!strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"NewsExtensions","no"),"yes");*/
	int		use_crosswords=!strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"CrossWords","no"),"yes");
	
	if(!where[0]){

		if(db->DBMode==UDM_DBMODE_CACHE){
			UdmClearCacheTree(Indexer->Conf);
		}

		if(use_crosswords&&db->DBMode!=UDM_DBMODE_CACHE){
			if((UDM_OK!=(rc=UdmDeleteAllFromCrossDict(Indexer,db))))return(rc);
		}
		if((UDM_OK!=(rc=UdmDeleteAllFromDict(Indexer,db))))return(rc);
		if((UDM_OK!=(rc=UdmDeleteAllFromUrl(Indexer,db))))return(rc);
	}else{
		j=0;
		while(1){
			size_t 		last = 0;
			char 		urlin[1024*3]="";
			char		limit[100]="";
			UDM_SQLRES	SQLres;
			
			if(db->DBSQL_LIMIT){
				sprintf(limit," LIMIT %d",URL_DELETE_CACHE);
			}
			sprintf(qbuf,"SELECT url.rec_id, url.url FROM url%s WHERE url.rec_id<>0 AND %s %s", db->from, where, limit);
			
			if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
				return rc;
			
			if(UdmSQLNumRows(&SQLres)){
				UDM_DOCUMENT Doc;
				
				bzero((void*)&Doc, sizeof(Doc));
				if(db->DBMode==UDM_DBMODE_CACHE){
					for(i=0;i<UdmSQLNumRows(&SQLres);i++){
						UdmVarListReplaceInt(&Doc.Sections,"ID",atoi(UdmSQLValue(&SQLres,i,0)));
						UdmVarListReplaceStr(&Doc.Sections, "URL", UdmSQLValue(&SQLres, i, 1));
						if(UDM_OK!=UdmDeleteURL(Indexer,&Doc,db))
							return(UDM_ERROR);
					}
					continue;
				}
				if(db->DBSQL_IN){
					urlin[0]=0;
					for(i=0;i<UdmSQLNumRows(&SQLres);i++){
						if(i)strcat(urlin,",");
						strcat(urlin,UdmSQLValue(&SQLres,i,0));
					}
					j+=i;
					switch(db->DBMode){
					case UDM_DBMODE_MULTI:
						for(i=MINDICT;i<MAXDICT;i++){
							if(last!=DICTNUM(i)){
								sprintf(qbuf,"DELETE FROM dict%d WHERE url_id in (%s)",DICTNUM(i),urlin);
								if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
									return rc;
								last=DICTNUM(i);
							}
						}
						break;
					case UDM_DBMODE_MULTI_CRC:
						for(i=MINDICT;i<MAXDICT;i++){
							if(last!=DICTNUM(i)){
								sprintf(qbuf,"DELETE FROM ndict%d WHERE url_id in (%s)",DICTNUM(i),urlin);
								if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
									return rc;
								last=DICTNUM(i);
							}
						}
						break;
					case UDM_DBMODE_CACHE:
						break;
					default:
						sprintf(qbuf,"DELETE FROM dict WHERE url_id in (%s)",urlin);
						if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
							return rc;
						break;
					}
					sprintf(qbuf,"DELETE FROM url WHERE rec_id in (%s)",urlin);
					if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
						return rc;
					
					sprintf(qbuf,"DELETE FROM urlinfo WHERE url_id in (%s)",urlin);
					if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
						return rc;

					sprintf(qbuf,"DELETE FROM links WHERE ot in (%s)",urlin);
					if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
						return rc;

					sprintf(qbuf,"DELETE FROM links WHERE k in (%s)",urlin);
					if(UDM_OK!=(rc=UdmSQLQuery(db,NULL,qbuf)))
						return rc;

					UdmSQLFree(&SQLres);
				}else{
					for(i=0;i<UdmSQLNumRows(&SQLres);i++)
						UdmVarListReplaceInt(&Doc.Sections,"ID",atoi(UdmSQLValue(&SQLres,i,0)));
						if(UDM_OK!=UdmDeleteURL(Indexer,&Doc,db))
							return(UDM_ERROR);
					j+=i;
				}
			}else{
				UdmSQLFree(&SQLres);
				break;
			}
		}
	}
	return(UDM_OK);
}

/********************* Categories ************************************/
static int UdmCatList(UDM_AGENT * Indexer,UDM_CATEGORY *Cat,UDM_DB *db){
	size_t		i, rows;
	char		qbuf[1024];
	UDM_SQLRES	SQLres;
	int		rc;
	
	if(db->DBType==UDM_DB_SAPDB){
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,lnk,name FROM categories WHERE path LIKE '%s__'",Cat->addr);
	}else{
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,link,name FROM categories WHERE path LIKE '%s__'",Cat->addr);
	}
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
		return rc;
	
	if( (rows = UdmSQLNumRows(&SQLres)) ){
		size_t nbytes;
		
		nbytes = sizeof(UDM_CATITEM) * (rows + Cat->ncategories);
		Cat->Category=(UDM_CATITEM*)realloc(Cat->Category,nbytes);
		for(i=0;i<rows;i++){
			UDM_CATITEM *r = &Cat->Category[Cat->ncategories];
			r[i].rec_id=atoi(UdmSQLValue(&SQLres,i,0));
			strcpy(r[i].path,UdmSQLValue(&SQLres,i,1));
			strcpy(r[i].link,UdmSQLValue(&SQLres,i,2));
			strcpy(r[i].name,UdmSQLValue(&SQLres,i,3));
		}
		Cat->ncategories+=rows;
	}
	UdmSQLFree(&SQLres);
	return UDM_OK;
}

static int UdmCatPath(UDM_AGENT *Indexer,UDM_CATEGORY *Cat,UDM_DB *db){
	size_t		i,l;
	char		qbuf[1024];
	int		rc;
	char            *head = NULL;
	
	l=(strlen(Cat->addr)/2)+1;
	Cat->Category=(UDM_CATITEM*)realloc(Cat->Category,sizeof(UDM_CATITEM)*(l+Cat->ncategories));
	head = (char *)malloc(2 * l + 1);

	if (head != NULL) {
		UDM_CATITEM	*r = &Cat->Category[Cat->ncategories];

	  for(i = 0; i < l; i++){
		UDM_SQLRES	SQLres;

		strncpy(head,Cat->addr,i*2);head[i*2]=0;

		if(db->DBType==UDM_DB_SAPDB){
			snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,lnk,name FROM categories WHERE path='%s'",head);
		}else{
			snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,link,name FROM categories WHERE path='%s'",head);
		}
		
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		
		if(UdmSQLNumRows(&SQLres)){
			r[i].rec_id=atoi(UdmSQLValue(&SQLres,0,0));
			strcpy(r[i].path,UdmSQLValue(&SQLres,0,1));
			strcpy(r[i].link,UdmSQLValue(&SQLres,0,2));
			strcpy(r[i].name,UdmSQLValue(&SQLres,0,3));
			Cat->ncategories++;
		}
		UdmSQLFree(&SQLres);
	  }
	  UDM_FREE(head);
	}
	return UDM_OK;
}


/******************* Search stuff ************************************/

static UDM_RESULT * UdmCloneList(UDM_AGENT * Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	UDM_RESULT	*Res;
	size_t		i;
	char		qbuf[256];
	UDM_SQLRES	SQLres;
	int		crc32=UdmVarListFindInt(&Doc->Sections,"crc32",0);
	int		origin_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
	int		rc;
	
	sprintf(qbuf,"SELECT rec_id,url,last_mod_time,docsize FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206) AND rec_id<>%d",crc32,origin_id);
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
		return(NULL);
	
	Res=UdmResultInit(NULL);
	Res->num_rows=UdmSQLNumRows(&SQLres);
	if(!(Res->num_rows)) {
	  UdmSQLFree(&SQLres);
	  return(Res);
	}
	if(Res->num_rows>5)Res->num_rows=5;
	
	Res->Doc=(UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));
	
	for(i=0;i<Res->num_rows;i++){
		time_t		last_mod_time;
		char		buf[64];
		UDM_DOCUMENT	*D=&Res->Doc[i];
		
		UdmDocInit(D);
		UdmVarListAddInt(&D->Sections,"ID",atoi(UdmSQLValue(&SQLres,i,0)));
		UdmVarListAddStr(&D->Sections,"URL",UdmSQLValue(&SQLres,i,1));
		last_mod_time=atol(UdmSQLValue(&SQLres,i,2));
		UdmTime_t2HttpStr(last_mod_time, buf);
		UdmVarListAddStr(&D->Sections,"Last-Modified",buf);
		UdmVarListAddInt(&D->Sections,"Content-Length",atoi(UdmSQLValue(&SQLres,i,3)));
		UdmVarListAddInt(&D->Sections,"crc32",crc32);
		UdmVarListAddInt(&D->Sections,"Origin-ID",origin_id);
	}
	UdmSQLFree(&SQLres);
	return(Res);
}

/********************************************************/

static int UdmLoadURLData(UDM_AGENT *A, UDM_RESULT *R) {
	UDM_SQLRES	SQLres;
	UDM_DB *db = A->Conf->db;
	size_t i,j;
	int rc;
	char qbuf[4*1024];

	if (R->CoordList.ncoords == 0) return UDM_OK;
	R->CoordList.Data = (UDM_URLDATA*)realloc(R->CoordList.Data, R->CoordList.ncoords * sizeof(UDM_URLDATA));
	if (db->DBSQL_IN) {
	  for (j = 0; j < R->CoordList.ncoords; j += 256) {
	    int notfirst = 0;
	    sprintf(qbuf, "SELECT rec_id, site_id, pop_rank FROM url WHERE rec_id IN (");
	    for (i = 0; (i < 256) && (j + i < R->CoordList.ncoords); i++) {
	      sprintf(UDM_STREND(qbuf), "%s%d", (notfirst) ? "," : "", R->CoordList.Coords[i].url_id);
	      notfirst = 1;
	    }
	    sprintf(UDM_STREND(qbuf), ") ORDER BY rec_id");
	    if (UDM_OK != (rc = UdmSQLQuery(db, &SQLres, qbuf)))
	      return rc;
	    for(i = 0; i < UdmSQLNumRows(&SQLres); i++){
	      R->CoordList.Data[i + j].url_id = atoi(UdmSQLValue(&SQLres, i, 0));
	      
	      if (R->CoordList.Data[j + i].url_id != R->CoordList.Coords[j + i].url_id)
		UdmLog(A, UDM_LOG_ERROR, "Crd url_id (%d) != Dat url_id (%d)", 
		       R->CoordList.Coords[j + i].url_id, R->CoordList.Data[j + i].url_id);
	      R->CoordList.Data[i + j].site_id = atoi(UdmSQLValue(&SQLres, i, 1));
	      R->CoordList.Data[i + j].pop_rank = atof(UdmSQLValue(&SQLres, i, 2));
	    }
	    UdmSQLFree(&SQLres);
	  }

	} else {
	  for (i = 0; i < R->CoordList.ncoords; i++) {
	    snprintf(qbuf, sizeof(qbuf), "SELECT site_id, pop_rank FROM url WHERE rec_id=%d", R->CoordList.Coords[i].url_id);
	    if (UDM_OK != (rc = UdmSQLQuery(db, &SQLres, qbuf)))
	      return rc;
	    if(UdmSQLNumRows(&SQLres)){
	      R->CoordList.Data[i].url_id = R->CoordList.Coords[i].url_id;
	      R->CoordList.Data[i].site_id = atoi(UdmSQLValue(&SQLres, 0, 0));
	      R->CoordList.Data[i].pop_rank = atof(UdmSQLValue(&SQLres, 0, 1));
	    }
	    UdmSQLFree(&SQLres);
	  }
	}
	return UDM_OK;
}

/********************************************************/

static int UdmSortAndGroupByURL(UDM_AGENT *A,UDM_RESULT *R){
	unsigned long	ticks=UdmStartTimer();
	int		use_site_id = ((!strcasecmp(UdmVarListFindStr(&A->Conf->Vars, "GroupBySite", "no"), "yes")) 
	  && (UdmVarListFindInt(&A->Conf->Vars, "site", 0) == 0));
	
	UdmLog(A,UDM_LOG_DEBUG,"Start sort by url_id %d words",R->CoordList.ncoords);
	UdmSortSearchWordsByURL(R->CoordList.Coords,R->CoordList.ncoords);
	ticks=UdmStartTimer()-ticks;
	UdmLog(A,UDM_LOG_DEBUG,"Stop sort by url_id:\t%.2f",(float)ticks/1000);
	
	UdmLog(A,UDM_LOG_DEBUG,"Start group by url_id %d docs",R->CoordList.ncoords);
	ticks=UdmStartTimer();
	UdmGroupByURL(A,R);
	ticks=UdmStartTimer()-ticks;
	UdmLog(A,UDM_LOG_DEBUG,"Stop group by url_id:\t%.2f",(float)ticks/1000);


	UdmLog(A,UDM_LOG_DEBUG,"Start load url data %d docs",R->CoordList.ncoords);
	ticks=UdmStartTimer();
	UdmLoadURLData(A, R);
	ticks=UdmStartTimer()-ticks;
	UdmLog(A,UDM_LOG_DEBUG,"Stop load url data:\t%.2f",(float)ticks/1000);

	if (use_site_id) {
	  UdmLog(A,UDM_LOG_DEBUG,"Start sort by site_id %d words",R->CoordList.ncoords);
	  UdmSortSearchWordsBySite(&R->CoordList, R->CoordList.ncoords);
	  ticks=UdmStartTimer()-ticks;
	  UdmLog(A,UDM_LOG_DEBUG,"Stop sort by site_id:\t%.2f",(float)ticks/1000);
	
	  UdmLog(A,UDM_LOG_DEBUG,"Start group by site_id %d docs", R->CoordList.ncoords);
	  ticks=UdmStartTimer();
	  UdmGroupBySite(A, R);
	  ticks=UdmStartTimer() - ticks;
	  UdmLog(A,UDM_LOG_DEBUG,"Stop group by site_id:\t%.2f", (float)ticks/1000);
	} 

	UdmLog(A,UDM_LOG_DEBUG,"Start sort by pop_rank %d words",R->CoordList.ncoords);
	UdmSortSearchWordsByPopRank(R, &R->CoordList, R->CoordList.ncoords);
	ticks=UdmStartTimer()-ticks;
	UdmLog(A,UDM_LOG_DEBUG,"Stop sort by pop_rank:\t%.2f",(float)ticks/1000);
	
	return UDM_OK;
}

static int UdmFindWordsBlob(UDM_AGENT *Agent, UDM_RESULT *Res,UDM_DB *db){
	
#ifdef HAVE_MYSQL_0
if(db->DBDriver==UDM_DB_MYSQL){
	size_t			wordnum;
	UDM_URLCRDLISTLIST	Tmp;
	size_t			totalcoords=0;
	size_t			i,nbytes;
	UDM_SQLRES		*sqlres;
	
	nbytes=sizeof(UDM_SQLRES)*Res->WWList.nwords;
	sqlres=(UDM_SQLRES*)malloc(nbytes);
	bzero((void*)sqlres, nbytes);
	bzero((void*)&Tmp, sizeof(Tmp));
	
	for(wordnum=0;wordnum<Res->WWList.nwords;wordnum++){
		int		crc32,tablenum,numrows,rc;
		char		escwrd[128];
		char		qbuf[128];
		UDM_WIDEWORD	*W=&Res->WWList.Word[wordnum];
		MYSQL_ROW	row;
		unsigned long	ticks=UdmStartTimer();
		
		if (W->origin == UDM_WORD_ORIGIN_STOP) continue;
		crc32=UdmStrCRC32(W->word);
		tablenum=(unsigned)crc32 % db->numtables;
		UdmDBEscStr(db->DBType,escwrd,W->word,strlen(W->word));
		
		sprintf(qbuf,"SELECT tstamp,coords FROM bdict%02X WHERE word='%s'",tablenum,W->word);
		UdmLog(Agent,UDM_LOG_DEBUG,"%s",qbuf);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&sqlres[wordnum],qbuf)))
			return rc;
		
		ticks=UdmStartTimer()-ticks;
		numrows=UdmSQLNumRows(&sqlres[wordnum]);
		UdmLog(Agent,UDM_LOG_DEBUG,"%d rows in result %.2f",numrows,(float)ticks/1000);
		
		ticks=UdmStartTimer();
		UdmLog(Agent,UDM_LOG_DEBUG,"Start fetching %d rows",numrows);
		
		while((row=mysql_fetch_row((sqlres[wordnum].mysqlres)))){
			unsigned long	*lengths=mysql_fetch_lengths(sqlres[wordnum].mysqlres);
			size_t		numcoords=lengths[1]/sizeof(UDM_URL_CRD);
			
			Tmp.List=(UDM_URLCRDLIST*)realloc(Tmp.List,(Tmp.nlists+1)*sizeof(Tmp.List[0]));
			Tmp.List[Tmp.nlists].Coords=(UDM_URL_CRD*)row[1];
			Tmp.List[Tmp.nlists].ncoords=numcoords;
			Tmp.List[Tmp.nlists].order = wordnum; /*Res->WWList.Word[wordnum].order;*/
			Tmp.nlists++;
			totalcoords+=numcoords;
		}
		
		ticks=UdmStartTimer()-ticks;
		UdmLog(Agent,UDM_LOG_DEBUG,"Done %d coords %.2f",totalcoords,(float)ticks/1000);
	}
	
	Res->CoordList.Coords=(UDM_URL_CRD*)realloc(Res->CoordList.Coords,(Res->CoordList.ncoords+totalcoords)*(sizeof(UDM_URL_CRD)));
	
	for(i=0;i<Tmp.nlists;i++){
		UDM_URL_CRD	*Cb=Res->CoordList.Coords+Res->CoordList.ncoords;
		UDM_URL_CRD	*Ce=Cb+Tmp.List[i].ncoords;
		size_t		order=Tmp.List[i].order;
		
		memcpy(Cb,Tmp.List[i].Coords,Tmp.List[i].ncoords*sizeof(UDM_URL_CRD));
		
		for(;Cb<Ce;Cb++){
			uint4	coord=Cb->coord;
			
			coord = coord & 0xFFFFFF00;
			Cb->coord = coord + (order & 0xFF);
		}
		Res->CoordList.ncoords+=Tmp.List[i].ncoords;
	}
	
	for(wordnum=0;wordnum<Res->WWList.nwords;wordnum++)
		UdmSQLFree(&sqlres[wordnum]);
	
	free(sqlres);
	free(Tmp.List);
	return UDM_OK;
}
#endif

#if defined HAVE_PGSQL || defined HAVE_IBASE
if(db->DBDriver==UDM_DB_PGSQL || db->DBDriver==UDM_DB_IBASE){
	size_t		wordnum,nbytes;
	UDM_SQLRES	*sqlres;
	size_t		totalcoords=0;
	int		rc;	
	
	nbytes=sizeof(UDM_SQLRES)*Res->WWList.nwords;
	sqlres=(UDM_SQLRES*)malloc(nbytes);
	bzero((void*)sqlres, nbytes);
	
	for(wordnum=0;wordnum<Res->WWList.nwords;wordnum++){
		int		crc32;
		size_t          tablenum, numrows, row;
		char		escwrd[128];
		char		qbuf[128];
		UDM_WIDEWORD	*W=&Res->WWList.Word[wordnum];
		unsigned long	ticks=UdmStartTimer();
		
		if (W->origin == UDM_WORD_ORIGIN_STOP) continue;
		crc32=UdmStrCRC32(W->word);
		tablenum=(unsigned)crc32 % db->numtables;
		UdmDBEscStr(db->DBType,escwrd,W->word,strlen(W->word));
		
		sprintf(qbuf,"SELECT tstamp,ncoords,coords FROM bdict%02X WHERE word='%s'",tablenum,W->word);
		UdmLog(Agent,UDM_LOG_DEBUG,"%s",qbuf);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&sqlres[wordnum],qbuf)))
			return rc;
		ticks=UdmStartTimer()-ticks;
		numrows=UdmSQLNumRows(&sqlres[wordnum]);
		UdmLog(Agent,UDM_LOG_DEBUG,"%d rows in result %.2f",numrows,(float)ticks/1000);
		
		ticks=UdmStartTimer();
		UdmLog(Agent,UDM_LOG_DEBUG,"Start fetching %d rows",numrows);
		
		for(row=0;row<numrows;row++){
			size_t	numcoords=atoi(UdmSQLValue(&sqlres[wordnum],row,1));
			totalcoords+=numcoords;
		}
		ticks=UdmStartTimer()-ticks;
		UdmLog(Agent,UDM_LOG_DEBUG,"Done %d coords %.2f",totalcoords,(float)ticks/1000);
	}
	
	Res->CoordList.Coords=(UDM_URL_CRD*)realloc(Res->CoordList.Coords,(Res->CoordList.ncoords+totalcoords)*(sizeof(UDM_URL_CRD)));
	
	for(wordnum=0;wordnum<Res->WWList.nwords;wordnum++){
		size_t	order = wordnum; /* Res->WWList.Word[wordnum].order; */
		size_t	row;
		size_t	numrows=UdmSQLNumRows(&sqlres[wordnum]);
		
		for(row=0;row<numrows;row++){
			size_t		numcoords=atoi(UdmSQLValue(&sqlres[wordnum],row,1));
			const char	*coords=UdmSQLValue(&sqlres[wordnum],row,2);
			size_t		ncoord=0;
			UDM_URL_CRD	*C=&Res->CoordList.Coords[Res->CoordList.ncoords];
			
			while(coords[0] && ncoord<numcoords){
				int	coord;
				
				while(coords[0]=='{' || coords[0]==',')coords++;
				C[ncoord].url_id=atoi(coords);
				while(coords[0]!='{' && coords[0]!=',' && coords[0])coords++;
				
				while(coords[0]=='{' || coords[0]==',')coords++;
				coord=atoi(coords);
				while(coords[0]!='{' && coords[0]!=',' && coords[0])coords++;
				
				coord = coord & 0xFFFFFF00;
				C[ncoord].coord = coord +(order & 0xFF);
				
				ncoord++;
			}
			
			/* 
			  Clear the tail if something is wrong: 
			  if bdict.ncoords doesn't correspond the
			  number of fetched word coordinates
			*/
			
			for(;ncoord<numcoords;ncoord++){
				C[ncoord].url_id=0;
				C[ncoord].coord=0;
			}
			
			Res->CoordList.ncoords+=numcoords;
		}
		UdmSQLFree(&sqlres[wordnum]);
	}
	free(sqlres);
	return UDM_OK;
}
#endif
	return UDM_OK;
}

static int UdmFindWordsSQL(UDM_AGENT * query,UDM_RESULT *Res,UDM_DB *db){
	char		qbuf[1024*4];
	size_t		wordnum;
	int		has_crosswrd=0;
	UDM_SQLRES	SQLres;
	int		word_match=UdmMatchMode(UdmVarListFindStr(&query->Conf->Vars,"wm","wrd"));
	int		wf[256];
	const char	*where=BuildWhere(query->Conf,db);
	int		use_crosswords=!strcasecmp(UdmVarListFindStr(&query->Conf->Vars,"CrossWords","no"),"yes");
	unsigned long 	ticks=UdmStartTimer();
	int		rc;
	
	UdmWeightFactorsInit(UdmVarListFindStr(&query->Conf->Vars,"wf",""),wf);
	
	/* Now find each word */
	for(wordnum=0;wordnum<Res->WWList.nwords;wordnum++){
		size_t	numrows,firstnum,curnum,tnum,tmin,tmax,tlst = (size_t)-1;
		size_t	i;
		char	tablename[32]="dict";
		char	escwrd[1000];
		UDM_WIDEWORD	*W=&Res->WWList.Word[wordnum];
		
		if (W->origin == UDM_WORD_ORIGIN_STOP) continue;
		
		UdmDBEscStr(db->DBType,escwrd,W->word,strlen(W->word));

		if((db->DBMode==UDM_DBMODE_MULTI)&&(word_match!=UDM_MATCH_FULL)){
			/* This is for substring search!  */
			/* In Multi mode: we have to scan */
			/* almost all tables except those */
			/* with to short words            */
			
			tmin=DICTNUM(strlen(W->word));
			tmax=MAXDICT;
		}else{
			tmin=tmax=DICTNUM(strlen(W->word));
		}

		for(tnum=tmin;tnum<=tmax;tnum++){

			if(tlst!=DICTNUM(tnum)){
				tlst=DICTNUM(tnum);		

				ticks=UdmStartTimer();
				UdmLog(query,UDM_LOG_DEBUG,"Start search for '%s'",Res->WWList.Word[wordnum].word);
				
				switch(db->DBMode){
				case UDM_DBMODE_MULTI:
					sprintf(tablename,"dict%d",DICTNUM(tnum));
					break;
				case UDM_DBMODE_MULTI_CRC:
					sprintf(tablename,"ndict%d",DICTNUM(tnum));
					break;
				case UDM_DBMODE_SINGLE_CRC:
					strcpy(tablename,"ndict");
					break;
				default:
					break;
				}
				if((db->DBMode==UDM_DBMODE_SINGLE_CRC)||(db->DBMode==UDM_DBMODE_MULTI_CRC)){
					int crc;
					crc=W->crcword;
					if(where[0]){
						snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s, url%s \
WHERE %s.word_id=%d \
AND url.rec_id=%s.url_id AND %s",
						tablename,tablename,
						tablename, db->from, tablename,
						crc,tablename,where);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"SELECT url_id,intag FROM %s WHERE %s.word_id=%d",
							 tablename,tablename,crc);
					}
				}else{
					char cmparg[256];
					switch(word_match){	
						case UDM_MATCH_BEGIN:
							sprintf(cmparg," LIKE '%s%%'",escwrd);
							break;
						case UDM_MATCH_END:
							sprintf(cmparg," LIKE '%%%s'",escwrd);
							break;
						case UDM_MATCH_SUBSTR:
							sprintf(cmparg," LIKE '%%%s%%'",escwrd);
							break;
						case UDM_MATCH_FULL:
						default:
							sprintf(cmparg,"='%s'",escwrd);
							break;
					}
					if(where[0]){
						snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s, url%s \
WHERE %s.word%s \
AND url.rec_id=%s.url_id AND %s",
						tablename,tablename,
						tablename, db->from, tablename,
						cmparg,tablename,where);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT url_id,intag FROM %s,url WHERE %s.word%s AND url.rec_id=%s.url_id",
							 tablename, tablename, cmparg, tablename);
					}
				}
				if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
					return rc;
					
				numrows=UdmSQLNumRows(&SQLres);
				ticks=UdmStartTimer()-ticks;
				UdmLog(query,UDM_LOG_DEBUG,"Stop search for '%s'\t%.2f  %d found",Res->WWList.Word[wordnum].word,(float)ticks/1000,numrows);
				
				/* Add new found word to the list */
				Res->CoordList.Coords = (Res->CoordList.ncoords+numrows) ? 
				  (UDM_URL_CRD*)UdmXrealloc(Res->CoordList.Coords,(Res->CoordList.ncoords+numrows)*sizeof(UDM_URL_CRD)) 
				  : NULL;
				
				firstnum=curnum=Res->CoordList.ncoords;
				for(i=0;i<numrows;i++){
					int url_id = 0;
					uint4 coord=0;
					uint4 weight;
					uint4 section;
					
					url_id=atoi(UdmSQLValue(&SQLres,i,0));
					coord=atoi(UdmSQLValue(&SQLres,i,1));
					
					section=UDM_WRDSEC(coord);
					weight=wf[section];
					if(weight){
						coord = coord & 0xFFFFFF00;
						Res->CoordList.Coords[curnum].url_id=url_id;
						Res->CoordList.Coords[curnum].coord = coord + (wordnum /* Res->WWList.Word[wordnum].order */ & 0xFF);
						curnum++;
					}
				}
				UdmSQLFree(&SQLres);
				Res->WWList.Word[wordnum].count+=curnum-firstnum;
				Res->CoordList.ncoords=curnum;
				Res->CoordList.Coords = (curnum) ? 
				  (UDM_URL_CRD*)UdmXrealloc(Res->CoordList.Coords, curnum * sizeof(UDM_URL_CRD))
				  : NULL;
			}
		}	
	}
	
	/* Now find each word in crosstable */
	has_crosswrd=((use_crosswords)&&(db->DBMode!=UDM_DBMODE_CACHE));
	for(wordnum=0;((has_crosswrd)&&(wordnum<Res->WWList.nwords));wordnum++){
		size_t		numrows,firstnum,curnum;
		char		tablename[32]="";
		char		escwrd[1000];
		size_t		i;
		UDM_WIDEWORD	*W=&Res->WWList.Word[wordnum];

		if(W->origin == UDM_WORD_ORIGIN_STOP) continue;

		UdmDBEscStr(db->DBType,escwrd,W->word,strlen(W->word));
		
		ticks=UdmStartTimer();
		UdmLog(query,UDM_LOG_DEBUG,"Start search for '%s'",W->word);
		
		switch(db->DBMode){
		case UDM_DBMODE_SINGLE_CRC:
		case UDM_DBMODE_MULTI_CRC:
			strcpy(tablename,"ncrossdict");
			break;
		case UDM_DBMODE_MULTI:
		default:
			strcpy(tablename,"crossdict");
			break;
		}
		if((db->DBMode==UDM_DBMODE_SINGLE_CRC)||
			(db->DBMode==UDM_DBMODE_MULTI_CRC)){
			int crc;
			crc=W->crcword;
			if(where[0]){
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s, url%s \
WHERE %s.word_id=%d \
AND url.rec_id=%s.url_id AND %s",
				tablename,tablename,
				tablename, db->from, tablename,
				crc,tablename,where);
			}else{
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT url_id,intag FROM %s,url WHERE %s.word_id=%d AND url.rec_id=%s.url_id", 
					 tablename, tablename, crc, tablename);
			}
		}else{
			char cmparg[256];
			switch(word_match){	
				case UDM_MATCH_BEGIN:
					sprintf(cmparg," LIKE '%s%%'",escwrd);
					break;
				case UDM_MATCH_END:
					sprintf(cmparg," LIKE '%%%s'",escwrd);
					break;
				case UDM_MATCH_SUBSTR:
					sprintf(cmparg," LIKE '%%%s%%'",escwrd);
					break;
				case UDM_MATCH_FULL:
				default:
					sprintf(cmparg,"='%s'",escwrd);
					break;
			}
			if(where[0]){
		
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s, url%s \
WHERE %s.word%s \
AND url.rec_id=%s.url_id AND %s",
				tablename,tablename,
				tablename, db->from, tablename,
				cmparg,tablename,where);
			}else{
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT url_id,intag FROM %s,url WHERE %s.word%s AND url.rec_id=%s.url_id",
					 tablename, tablename, cmparg, tablename);
			}
		}
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		
		numrows=UdmSQLNumRows(&SQLres);
		ticks=UdmStartTimer()-ticks;
		UdmLog(query,UDM_LOG_DEBUG,"Stop search for '%s'\t%.2f %d found",Res->WWList.Word[wordnum].word,(float)ticks/1000,numrows);
		
		/* Add new found word to the list */
		Res->CoordList.Coords=(UDM_URL_CRD*)UdmXrealloc(Res->CoordList.Coords,(Res->CoordList.ncoords+numrows)*sizeof(UDM_URL_CRD));
		
		firstnum=curnum=Res->CoordList.ncoords;
		for(i=0;i<numrows;i++){
			int url_id=0;
			int coord=0;
			int weight;
			int section;
			
			url_id=atoi(UdmSQLValue(&SQLres,i,0));
			coord=atoi(UdmSQLValue(&SQLres,i,1));
			
			section=UDM_WRDSEC(coord);
			weight=wf[section];
			if(weight){
				coord = coord & 0xFFFFFF00;
				Res->CoordList.Coords[curnum].url_id=url_id;
				Res->CoordList.Coords[curnum].coord = coord + (wordnum /* Res->WWList.Word[wordnum].order */ & 0xFF);
				curnum++;
			}
		}
		UdmSQLFree(&SQLres);
		Res->CoordList.ncoords=curnum;
		Res->WWList.Word[wordnum].count+=curnum-firstnum;
		Res->CoordList.Coords=(UDM_URL_CRD*)UdmXrealloc(Res->CoordList.Coords,curnum*sizeof(UDM_URL_CRD));
	}
	return UDM_OK;
}


static int UdmTrack(UDM_AGENT * query,UDM_RESULT *Res,UDM_DB *db){
	char		*qbuf;
	char		*text_escaped;
	const char	*words=UdmVarListFindStr(&query->Conf->Vars,"q",""); /* "q-lc" qas here */
	size_t          escaped_len, qbuf_len;
	int             res;
	
	if (db->DBDriver == UDM_DB_FILES)
		return UDM_OK;

	if (*words == '\0') return UDM_OK; /* do not store empty queries */

	escaped_len = 4 * strlen(words);
	qbuf_len = escaped_len + 4096;

	if ((qbuf = (char*)malloc(qbuf_len)) == NULL) return UDM_ERROR;
	if ((text_escaped = (char*)malloc(escaped_len)) == NULL) { free(qbuf); return UDM_ERROR; }
	
	/* Escape text to track it  */
	UdmDBEscStr(db->DBType, text_escaped, words, strlen(words));
	
	snprintf(qbuf, qbuf_len - 1, "INSERT INTO qtrack (ip,qwords,qtime,found,ps,np,m,wm,o,t,cat,ul,wf,g,tmplt,groupbysite,site_id) VALUES ('%s','%s',%d,%d,%d,%d,'%s','%s',%d,'%s','%s','%s','%s','%s','%s','%c',%d)",
		UdmVarListFindStr(&query->Conf->Vars, "IP", ""),
		text_escaped,(int)time(NULL),Res->total_found,
		UdmVarListFindInt(&query->Conf->Vars, "ps", 20),
		UdmVarListFindInt(&query->Conf->Vars, "np", 0),
		UdmVarListFindStr(&query->Conf->Vars, "orig_m", ""),
		UdmVarListFindStr(&query->Conf->Vars, "wm", ""),
		UdmVarListFindInt(&query->Conf->Vars, "o", 0),
		UdmVarListFindStr(&query->Conf->Vars, "t", ""),
		UdmVarListFindStr(&query->Conf->Vars, "cat", ""),
		UdmVarListFindStr(&query->Conf->Vars, "ul", ""),
		UdmVarListFindStr(&query->Conf->Vars, "wf", ""),
		UdmVarListFindStr(&query->Conf->Vars, "g", ""),
		UdmVarListFindStr(&query->Conf->Vars, "tmplt", ""),
		toupper(*(UdmVarListFindStr(&query->Conf->Vars, "GroupBySite", "No"))),
		UdmVarListFindInt(&query->Conf->Vars, "site", 0) );
	
#ifdef DEBUG_SQL
	fprintf(stderr, "TrackQuery: %s\n", qbuf);
#endif
	res = UdmSQLQuery(db, NULL, qbuf);
	free(text_escaped);
	free(qbuf);
	return res;
}

static void SQLResToDoc(UDM_ENV *Conf, UDM_DOCUMENT *D, UDM_SQLRES *sqlres, size_t i){
	time_t		last_mod_time;
	char		dbuf[128];
	const char	*format = UdmVarListFindStrTxt(&Conf->Vars, "DateFormat", "%a, %d %b %Y, %X %Z");
	
	UdmVarListReplaceStr(&D->Sections,"URL",UdmSQLValue(sqlres,i,1));
	last_mod_time=atol(UdmSQLValue(sqlres,i,2));
	if (strftime(dbuf, 128, format, localtime(&last_mod_time)) == 0) {
	  UdmTime_t2HttpStr(last_mod_time, dbuf);
	}
	UdmVarListReplaceStr(&D->Sections,"Last-Modified",dbuf);
	UdmVarListReplaceStr(&D->Sections,"Content-Length",UdmSQLValue(sqlres,i,3));
	last_mod_time=atol(UdmSQLValue(sqlres,i,4));
	if (strftime(dbuf, 128, format, localtime(&last_mod_time)) == 0) {
	  UdmTime_t2HttpStr(last_mod_time, dbuf);
	}
	UdmVarListReplaceStr(&D->Sections,"Next-Index-Time",dbuf);
	UdmVarListReplaceInt(&D->Sections,"Referrer-ID",atoi(UdmSQLValue(sqlres,i,5)));
	UdmVarListReplaceInt(&D->Sections,"crc32",atoi(UdmSQLValue(sqlres,i,6)));
	UdmVarListReplaceStr(&D->Sections, "Site_id", UdmSQLValue(sqlres, i, 7));
	UdmVarListReplaceStr(&D->Sections, "Pop_Rank", UdmSQLValue(sqlres, i, 8));
}


static int UdmResAddDocInfo(UDM_AGENT *query, UDM_RESULT *Res,UDM_DB *db){
	size_t		i;
	char		instr[1024*4]="";
	char		qbuf[1024*4];
	UDM_SQLRES	SQLres;
	int		rc;
	
	if(!Res->num_rows)return UDM_OK;
	
	if(db->DBSQL_IN){
		size_t	j;
		
		/* Compose IN string and set to zero url_id field */
		for(i=0;i<Res->num_rows;i++){
			sprintf(UDM_STREND(instr),"%s%d",i?",":"",
				UdmVarListFindInt(&Res->Doc[i].Sections,"ID",0));
		}
		
		snprintf(qbuf,sizeof(qbuf),"SELECT rec_id,url,last_mod_time,docsize,next_index_time,referrer,crc32,site_id,pop_rank FROM url WHERE rec_id IN (%s) ORDER BY rec_id",instr);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		
		for(j=0;j<Res->num_rows;j++) {
			for(i=0;i<UdmSQLNumRows(&SQLres);i++) {
				UDM_DOCUMENT	*D=&Res->Doc[j];
				int		url_id=UdmVarListFindInt(&D->Sections,"ID",0);
				
				if(url_id == atoi(UdmSQLValue(&SQLres,i,0))) {
					SQLResToDoc(query->Conf, D, &SQLres, i);
					break;
				}
			}
		}
		UdmSQLFree(&SQLres);
	}else{
		for(i=0;i<Res->num_rows;i++){
			UDM_DOCUMENT	*D=&Res->Doc[i];
			int		url_id=UdmVarListFindInt(&D->Sections,"ID",0);
			
			sprintf(qbuf,"SELECT rec_id,url,last_mod_time,docsize,next_index_time,referrer,crc32,site_id,pop_rank FROM url WHERE rec_id=%d",url_id);
			if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
				return rc;
			
			if(UdmSQLNumRows(&SQLres)){
				SQLResToDoc(query->Conf, D, &SQLres, 0);
			}
			UdmSQLFree(&SQLres);
		}
	}
	
	if(db->DBSQL_IN){
		size_t j;
		
		instr[0]='\0';
		/* Compose IN string and set to zero url_id field */
		for(i=0;i<Res->num_rows;i++){
			sprintf(UDM_STREND(instr),"%s%d",i?",":"",
				UdmVarListFindInt(&Res->Doc[i].Sections,"ID",0));
		}
		
		snprintf(qbuf,sizeof(qbuf),"SELECT url_id,sname,sval FROM urlinfo WHERE url_id IN (%s) ORDER BY url_id",instr);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		
		for(j=0;j<Res->num_rows;j++) {
			UDM_DOCUMENT	*D=&Res->Doc[j];
			int		url_id=UdmVarListFindInt(&D->Sections,"ID",0);
			
			for(i=0;i<UdmSQLNumRows(&SQLres);i++) {
				if(url_id == atoi(UdmSQLValue(&SQLres,i,0))){
					const char *sname=UdmSQLValue(&SQLres,i,1);
					const char *sval=UdmSQLValue(&SQLres,i,2);
					UdmVarListAddStr(&D->Sections,sname,sval);
				}
			}
		}
		UdmSQLFree(&SQLres);
	}else{
		for(i=0;i<Res->num_rows;i++){
			UDM_DOCUMENT	*D=&Res->Doc[i];
			size_t		row;
			int		url_id=UdmVarListFindInt(&D->Sections,"ID",0);
			
			sprintf(qbuf,"SELECT url_id,sname,sval FROM urlinfo WHERE url_id=%d",url_id);
			if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
				return rc;
			
			for(row=0;row<UdmSQLNumRows(&SQLres);row++) {
				const char *sname=UdmSQLValue(&SQLres,row,1);
				const char *sval=UdmSQLValue(&SQLres,row,2);
				UdmVarListAddStr(&D->Sections,sname,sval);
			}
			UdmSQLFree(&SQLres);
		}
	}
	return(UDM_OK);
}


static int UdmFindWords(UDM_AGENT * query,UDM_RESULT *Res,UDM_DB *db){
	const char	*cache_mode=NULL;
/*	int		page_size  =UdmVarListFindInt(&query->Conf->Vars,"ps",20);
	int		page_number=UdmVarListFindInt(&query->Conf->Vars,"np",0);
	size_t		topcount;*/
	unsigned long	ticks=UdmStartTimer();
	
	cache_mode  = UdmVarListFindStr(&query->Conf->Vars, "Cache", "no");
	UdmVarListAddStr(&query->Conf->Vars, "orig_m", UdmVarListFindStr(&query->Conf->Vars, "m", "all"));
	
	if( strcasecmp(cache_mode,"yes") || UdmSearchCacheFind(query,Res) ){
		/* If not found in search cache       */
		/* Let's do actual search now:        */
		/* Get groupped by url_id words array */
	
		if(db->DBMode==UDM_DBMODE_CACHE){
			UdmFindCache(query,Res);
		}else
		if(db->DBMode==UDM_DBMODE_BLOB){
			int res=UdmFindWordsBlob(query,Res,db);
			if(res!=UDM_OK)return res;
			UdmSortAndGroupByURL(query,Res);
		}else{
			int res=UdmFindWordsSQL(query,Res,db);
			if(res!=UDM_OK)return res;
			UdmSortAndGroupByURL(query,Res);
		}
		UdmLog(query,UDM_LOG_DEBUG,"Start sort by weight %d docs",Res->CoordList.ncoords);
		ticks=UdmStartTimer();
		
		/* Now sort results by the weight */
/*		
		topcount=page_size*(page_number+1)-1;
		if(topcount>=Res->CoordList.ncoords)topcount=Res->CoordList.ncoords-1;
		
		if(topcount<UDM_FAST_PRESORT_DOCS){
			UdmWrdTopSort(Res->CoordList.Coords,Res->CoordList.ncoords,topcount);
		}else{
			UdmSortSearchWordsByWeight(Res->CoordList.Coords,Res->CoordList.ncoords);
		}
		ticks=UdmStartTimer()-ticks;
		UdmLog(query,UDM_LOG_DEBUG,"Stop sort by weight:\t%.2f",(float)ticks/1000);
*/		
		/* Store into search result cache */
#if (WIN32|WINNT)
#else
		if((!strcasecmp(cache_mode,"yes"))&&(search_cache_size>-1)){
			fflush(stdout);
			fflush(stderr);
/*			if(fork()==0){*/
				/* Store to cache and exit program */
				UdmSearchCacheStore(query,Res);
/*				exit(0);
			}*/
		}
#endif
/*		Res->offset=1;*/
	}
	if(!strcasecmp(UdmVarListFindStr(&query->Conf->Vars, "TrackQuery", "no"),"yes")){
		UdmTrack(query,Res,db);
	}
	
	return UDM_OK;
}


int UdmFindSQL(UDM_AGENT *query,UDM_RESULT *Res){
	UDM_DB		*db=query->Conf->db;
	size_t		i;
	int		res=UDM_OK;
	int		page_size   = UdmVarListFindInt(&query->Conf->Vars, "ps", 20);
	int		page_number = UdmVarListFindInt(&query->Conf->Vars, "np", 0);
	unsigned long	ticks=UdmStartTimer();
	UDM_CHARSET	*lcs=UdmGetCharSet(UdmVarListFindStr(&query->Conf->Vars,"LocalCharset","iso-8859-1"));
	UDM_CHARSET	*bcs=UdmGetCharSet(UdmVarListFindStr(&query->Conf->Vars,"BrowserCharset","iso-8859-1"));
	char str[128];
	
	if(!lcs)lcs=UdmGetCharSet("iso-8859-1");
	if(!bcs)bcs=UdmGetCharSet("iso-8859-1");
	
	UdmLog(query,UDM_LOG_DEBUG,"Start words stuff");	
	
	UdmPrepare(query,Res);				/* Prepare query    */
	AutoWild(query,&query->Conf->Vars);
	if(UDM_OK!=(res=UdmFindWords(query,Res,db)))	/* Load words array */
		goto ret;
	
	/* Return if no words found   */
	if(!Res->CoordList.ncoords)
		goto ret;
	
	/* Find curent page offsets */
	/* And If there was a request for */
	/* too large page number we will  */
	/* display the only one last doc  */
	
	Res->total_found=Res->CoordList.ncoords;
	Res->first=page_number*page_size;	
	if(Res->first>=Res->total_found)Res->first=Res->total_found-1;
	
	/* If results more than 1 page */
	/* we must cut the tail        */
	if((Res->first+page_size)>Res->total_found){
		Res->num_rows=Res->total_found-Res->first;
	}else{
		Res->num_rows=page_size;
	}
	Res->last=Res->first+Res->num_rows-1;
	
	UdmLog(query,UDM_LOG_DEBUG,"First: %d TotalFound: %d NumRows: %d",Res->first,Res->total_found,Res->num_rows);
	ticks=UdmStartTimer()-ticks;
	UdmLog(query,UDM_LOG_DEBUG,"Stop words stuff: %.2f",(float)ticks/1000);
	
	ticks=UdmStartTimer();
	UdmLog(query,UDM_LOG_DEBUG,"Start URLs");
	
	/* Allocate an array for documents information */
	Res->Doc=(UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));
	
	/* Copy url_id and coord to result */
	for(i=0;i<Res->num_rows;i++){
		UdmDocInit(&Res->Doc[i]);
		UdmVarListAddInt(&Res->Doc[i].Sections,"ID",Res->CoordList.Coords[i + Res->first /* * Res->offset*/].url_id);
		snprintf(str, 128, "%.3f%%", ((double)Res->CoordList.Coords[i + Res->first /* * Res->offset*/].coord) / 1000);
		UdmVarListReplaceStr(&Res->Doc[i].Sections, "Score", str );
	}
	
	/* Load documents information */
	if(UDM_OK!=(res=UdmResAddDocInfo(query,Res,db)))
		goto ret;
	
	ticks=UdmStartTimer()-ticks;
	UdmLog(query,UDM_LOG_DEBUG,"Stop  URLs: %.2f",(float)ticks/1000);
	
	
	ticks=UdmStartTimer();
	UdmLog(query,UDM_LOG_DEBUG,"Start Clones");
	
	if(!strcasecmp(UdmVarListFindStr(&query->Conf->Vars,"DetectClones","yes"),"yes")){
		size_t num=Res->num_rows;
		for(i=0;i<num;i++){
			UDM_RESULT *Cl=UdmCloneList(query,&Res->Doc[i],db);
			if(Cl){
				Res->Doc=(UDM_DOCUMENT*)realloc(Res->Doc,sizeof(UDM_DOCUMENT)*(Res->num_rows+Cl->num_rows));
				memcpy(&Res->Doc[Res->num_rows],Cl->Doc,sizeof(UDM_DOCUMENT)*Cl->num_rows);
				Res->num_rows+=Cl->num_rows;
				UDM_FREE(Cl->Doc);
				UdmResultFree(Cl);
			}
		}
	}
	ticks=UdmStartTimer()-ticks;
	UdmLog(query,UDM_LOG_DEBUG,"Stop  Clones: %.2f",(float)ticks/1000);
	
	/* first and last begins from 0, make it begin from 1 */
	Res->first++;
	Res->last++;

	for(i=0;i<Res->num_rows;i++){
		UdmVarListReplaceInt(&Res->Doc[i].Sections,"Order",(int)(Res->first+i));
	}
	
ret:	
	strcpy(query->Conf->errstr,db->errstr);
	return res;
}

int UdmURLDataSQL(UDM_ENV *Conf, UDM_URLDATALIST *L) {
	UDM_SQLRES	SQLres;
	UDM_DB		*db=Conf->db;
	size_t		i;
	int		rc=UDM_OK;
	
	L->nitems=0;
	L->Item=NULL;
	
	if (UDM_OK != (rc=UdmSQLQuery(db, &SQLres, "SELECT rec_id,site_id,pop_rank FROM url ORDER by rec_id")))
		return rc;
	
	L->nitems = UdmSQLNumRows(&SQLres);
	L->Item = (UDM_URLDATA*)malloc(L->nitems*sizeof(UDM_URLDATA));
	if (L->Item == NULL) {
		L->nitems = 0;
		rc=UDM_ERROR;
		goto freeex;
	}
	for (i = 0; i < L->nitems; i++) {
		L->Item[i].url_id = atoi(UdmSQLValue(&SQLres, i, 0));
		L->Item[i].site_id = atoi(UdmSQLValue(&SQLres, i, 1));
		L->Item[i].pop_rank = atof(UdmSQLValue(&SQLres, i, 2));
	}
freeex:
	UdmSQLFree(&SQLres);
	return rc;
}



/***********************************************************/
/*  HTDB stuff:  Indexing of database content              */
/***********************************************************/


static char * get_path_part(char *path,char *dst,int part){
	const char *s;
	char *e;
	int i=0;

	s=path;
	while(*s){
		if(i==part){
			if((e=strchr(s,'/')))strncpy(dst,s,(unsigned)(e-s));
			else	strcpy(dst,s);
			return(dst);
		}
		if(*s=='/')i++;
		s++;
	}
	*dst=0;
	return(dst);
}
static char * include_params(const char *src,char *path,char *dst){
	const char *s;
	char *e;
	int i;
	
	s=src;e=dst;*e=0;
	while(*s){
		if((*s=='\\')){
			*e=*(s+1);e++;*e=0;s+=2;
			continue;
		}
		if(*s!='$'){
			*e=*s;
			s++;e++;*e=0;
			continue;
		}
		s++;i=atoi(s);
		while((*s>='0')&&(*s<='9'))s++;
		get_path_part(path,e,i);
		while(*e)e++;
	}
	return(dst);
}


static int UdmHTDBGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_DB *db){
	char		qbuf[1024*4]="";
	char		*end=Doc->Buf.buf;
	UDM_SQLRES	SQLres;
	UDM_URL		realURL;
	const char	*url=UdmVarListFindStr(&Doc->Sections,"URL","");
	const char	*htdblist=UdmVarListFindStr(&Doc->Sections,"HTDBList","");
	const char	*htdbdoc=UdmVarListFindStr(&Doc->Sections,"HTDBDoc","");
	int		rc;
	
	Doc->Buf.buf[0]=0;
	UdmURLParse(&realURL,url);
	
	if(realURL.filename[0]){
		char real_path[1024]="";
		
		snprintf(real_path,sizeof(real_path)-1,"%s%s",realURL.path,realURL.filename);
		real_path[sizeof(real_path)-1]='\0';
		include_params(htdbdoc,real_path,qbuf);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		if(UdmSQLNumRows(&SQLres)==1){
			strcpy(Doc->Buf.buf,UdmSQLValue(&SQLres,0,0));
		}else{
			sprintf(Doc->Buf.buf,"HTTP/1.0 404 Not Found\r\n\r\n");
		}
		UdmSQLFree(&SQLres);
	}else{
		size_t	i;
		int	url_id=UdmVarListFindInt(&Doc->Sections,"ID",0);
		int	hops=UdmVarListFindInt(&Doc->Sections,"Hops",0);
		
		include_params(htdblist,realURL.path,qbuf);
		if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf)))
			return rc;
		sprintf(Doc->Buf.buf,"HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n<HTML><BODY>\n");
		end=UDM_STREND(Doc->Buf.buf);
		
		for(i=0;i<UdmSQLNumRows(&SQLres);i++){
			UDM_HREF Href;
			UdmHrefInit(&Href);
			Href.referrer=url_id;
			Href.hops=hops+1;
			Href.url=strdup(UdmSQLValue(&SQLres,i,0));
			Href.method=UDM_METHOD_GET;
			UdmHrefListAdd(&Doc->Hrefs,&Href);
			free(Href.url);
		}
		UdmSQLFree(&SQLres);
		strcpy(end,"</BODY></HTML>\n");
	}
	end = UDM_STREND(Doc->Buf.buf);
	Doc->Buf.size=end-Doc->Buf.buf;
	return UDM_OK;
}

/************************** make index stuff *******************************/

static char *BuildLimitQuery(const char * field) {
  char qbuf[2048];
  char small[128];

  snprintf(small, 128, ":%s:", field);
  if (strstr(":status:docsize:next_index_time:last_mod_time:crc32:referrer:hops:seed:bad_since_time:site_id:pop_rank:url:", 
	     small) != NULL) {
    snprintf(qbuf, 2048, "SELECT %s,rec_id FROM url", field);
  } else if(strstr(":category:tag:", small) != NULL) {
    snprintf(qbuf, 2048, "SELECT s.%s,u.rec_id FROM server s, url u WHERE s.rec_id=u.site_id", field);
  } else {
    snprintf(qbuf, 2048, "SELECT i.sval,u.rec_id FROM url u,urlinfo i WHERE u.rec_id=i.url_id AND i.sname='%s'", field);
  }
  return strdup(qbuf);
}

int UdmLimit8SQL(UDM_ENV *Conf, UDM_UINT8URLIDLIST *L,const char *field, int type, UDM_DB *db){
	char		*qbuf = BuildLimitQuery(field);
	size_t		i;
	UDM_SQLRES	SQLres;
	int		rc=UDM_OK;
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf))) {
	        free(qbuf);
		return UDM_ERROR;
	}
	free(qbuf);
	L->nitems=UdmSQLNumRows(&SQLres);
	
	L->Item=(UDM_UINT8URLID*)malloc(L->nitems*sizeof(UDM_UINT8URLID));
	if(!L->Item){
		sprintf(db->errstr,"Error: %s",strerror(errno));
		db->errcode=1;
		UdmSQLFree(&SQLres);
		return UDM_ERROR;
	}
	for(i=0;i<L->nitems;i++){
		const char *val0=UdmSQLValue(&SQLres,i,0);
		const char *val1=UdmSQLValue(&SQLres,i,1);

		switch(type){
			case UDM_IFIELD_TYPE_HEX8STR: UdmDecodeHex8Str(val0,&L->Item[i].hi, &L->Item[i].lo, NULL, NULL); break;
			case UDM_IFIELD_TYPE_INT: L->Item[i].hi = atoi(val0); L->Item[i].lo = 0; break;
		}
		L->Item[i].url_id=atoi(val1);
	}
	UdmSQLFree(&SQLres);
	return rc;
}

int UdmLimit4SQL(UDM_ENV *Conf, UDM_UINT4URLIDLIST *L,const char *field, int type, UDM_DB *db){
	char		*qbuf = BuildLimitQuery(field);
	size_t		i;
	UDM_SQLRES	SQLres;
	int		rc=UDM_OK;
	
	if(UDM_OK!=(rc=UdmSQLQuery(db,&SQLres,qbuf))) {
	        free(qbuf);
		return rc;
	}
	free(qbuf);
	
	L->nitems=UdmSQLNumRows(&SQLres);
	
	L->Item=(UDM_UINT4URLID*)malloc(L->nitems*sizeof(UDM_UINT4URLID));
	if(!L->Item){
		sprintf(db->errstr,"Error: %s",strerror(errno));
		db->errcode=0;
		UdmSQLFree(&SQLres);
		return UDM_OK;
	}
	for(i=0;i<L->nitems;i++){
		const char *val0=UdmSQLValue(&SQLres,i,0);
		const char *val1=UdmSQLValue(&SQLres,i,1);

		switch(type){
			case UDM_IFIELD_TYPE_HOUR: L->Item[i].val = atoi(val0) / 3600; break;
			case UDM_IFIELD_TYPE_MIN: L->Item[i].val = atoi(val0) / 60; break;
			case UDM_IFIELD_TYPE_HOSTNAME: {
				UDM_URL url;
				if(!UdmURLParse(&url,val0)){
					if(url.hostname) L->Item[i].val=UdmStrCRC32(url.hostname);
					else L->Item[i].val=0;
				}else
					L->Item[i].val=0;
				break;
			case UDM_IFIELD_TYPE_STRCRC32: L->Item[i].val = UdmStrCRC32(val0); break;
			case UDM_IFIELD_TYPE_INT: L->Item[i].val = atoi(val0); break;
			}
		}
		L->Item[i].url_id=atoi(val1);
	}
	UdmSQLFree(&SQLres);
	return rc;
}


/***************************************************************************/

static int UdmPopRankCalculate(UDM_AGENT *A, UDM_DB *db) {
	UDM_SQLRES	SQLres, Res;
	char		qbuf[128];
	int             rc = UDM_OK;
	size_t		i, nrows;
	int		skip_same_site = !strcasecmp(UdmVarListFindStr(&A->Conf->Vars, "PopRankSkipSameSite","no"),"yes");
	int		feed_back = !strcasecmp(UdmVarListFindStr(&A->Conf->Vars, "PopRankFeedBack", "no"), "yes");


	if (feed_back) {
	  UdmLog(A, UDM_LOG_EXTRA, "calculate feed back servers weights");

	  if(UDM_OK != (rc = UdmSQLQuery(db, &SQLres, "SELECT site_id, SUM(pop_rank) FROM url GROUP BY site_id")))
	    goto Calc_unlock;
	  nrows = UdmSQLNumRows(&SQLres);

	  for (i = 0; i < nrows; i++) {
	    snprintf(qbuf, 128, "UPDATE server SET weight=%s WHERE rec_id=%s", UdmSQLValue(&SQLres, i, 1), UdmSQLValue(&SQLres, i, 0));
	    UdmSQLQuery(db, NULL, qbuf);
	  }
	  UdmSQLFree(&SQLres);
	  UdmSQLQuery(db, NULL, "UPDATE server SET weight=1 WHERE weight<1");
	}


	if(UDM_OK != (rc = UdmSQLQuery(db, &SQLres, "SELECT rec_id, url, weight FROM server WHERE command='S'")))
		goto Calc_unlock;
	
	nrows = UdmSQLNumRows(&SQLres);

	for (i = 0; i < nrows; i++) {

	  snprintf(qbuf, 128, "SELECT COUNT(*) FROM url WHERE site_id=%s", UdmSQLValue(&SQLres, i, 0));
	  if(UDM_OK != (rc = UdmSQLQuery(db, &Res, qbuf)))
		goto Calc_unlock;
	  UdmLog(A, UDM_LOG_EXTRA, "Site: %s Weight: %s URL count: %s",
		 UdmSQLValue(&SQLres, i, 0), UdmSQLValue(&SQLres, i, 2), UdmSQLValue(&Res, 0, 0)); 
	  if (atoi(UdmSQLValue(&Res, 0, 0)) > 0) {
	    snprintf(qbuf, 128, "UPDATE server SET pop_weight=(%s/%s.0) WHERE rec_id=%s",
		  UdmSQLValue(&SQLres, i, 2), UdmSQLValue(&Res, 0, 0), UdmSQLValue(&SQLres, i, 0));
	    UdmSQLQuery(db, NULL, qbuf);
	  }
	  UdmSQLFree(&Res);

	}
	UdmSQLFree(&SQLres);

	UdmLog(A, UDM_LOG_EXTRA, "update link weights");



	if (skip_same_site) {
	  if(UDM_OK != (rc = UdmSQLQuery(db, &SQLres, "SELECT ot, count(*) FROM links l, url uo, url uk WHERE uo.rec_id=l.ot AND uk.rec_id=l.k AND uo.site_id <> uk.site_id GROUP BY ot")))
		goto Calc_unlock;
	  UdmLog(A, UDM_LOG_DEBUG, "info: will skip links from same site");
	} else {
	  if(UDM_OK != (rc = UdmSQLQuery(db, &SQLres, "SELECT ot, count(*) FROM links GROUP BY ot")))
		goto Calc_unlock;
	}
	
	nrows = UdmSQLNumRows(&SQLres);

	for (i = 0; i < nrows; i++) {
	    snprintf(qbuf, 128, "SELECT s.pop_weight FROM server s, url u WHERE u.rec_id=%s AND s.rec_id=u.site_id",
		     UdmSQLValue(&SQLres, i, 0));
	    if(UDM_OK != (rc = UdmSQLQuery(db, &Res, qbuf)))
	      goto Calc_unlock;
	    if (UdmSQLNumRows(&Res)) {
	      snprintf(qbuf, 128, "UPDATE links SET weight = (%s/%s) WHERE ot=%s",
		       UdmSQLValue(&Res, 0, 0), UdmSQLValue(&SQLres, i, 1), UdmSQLValue(&SQLres, i, 0));
	      UdmSQLQuery(db, NULL, qbuf);
	    }
	    UdmSQLFree(&Res);
	}
	UdmSQLFree(&SQLres);

	UdmLog(A, UDM_LOG_EXTRA, "update page weights");





	if(UDM_OK != (rc = UdmSQLQuery(db, &SQLres, "SELECT SUM(weight),k  FROM links GROUP BY k")))
		goto Calc_unlock;

	nrows = UdmSQLNumRows(&SQLres);

	for (i = 0; i < nrows; i++) {
	  snprintf(qbuf, 128, "UPDATE url SET pop_rank=%s WHERE rec_id=%s", UdmSQLValue(&SQLres, i, 0), UdmSQLValue(&SQLres, i, 1) );
	  UdmSQLQuery(db, NULL, qbuf);
	}
	UdmSQLFree(&SQLres);
	rc = UDM_OK;

Calc_unlock:
	UdmLog(A, UDM_LOG_EXTRA, "Popularity rank done.");
	return rc;
}



int UdmURLActionSQL(UDM_AGENT * A, UDM_DOCUMENT * D, int cmd,UDM_DB *db){
	int res;

	switch(cmd){
		case UDM_URL_ACTION_DELETE:
			res=UdmDeleteURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_ADD:
			res=UdmAddURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_SUPDATE:
			res=UdmUpdateUrl(A,D,db);
			break;
			
		case UDM_URL_ACTION_LUPDATE:
			res=UdmLongUpdateURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_INSWORDS:
			res=UdmStoreWords(A,D,db);
			break;
			
		case UDM_URL_ACTION_INSCWORDS:
			res=UdmStoreCrossWords(A,D,db);
			break;
			
		case UDM_URL_ACTION_DELWORDS:
			res=UdmDeleteWordFromURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_DELCWORDS:
			res=UdmDeleteCrossWordFromURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_UPDCLONE:
			res=UdmUpdateClone(A,D,db);
			break;
			
		case UDM_URL_ACTION_REGCHILD:
			res=UdmRegisterChild(A,D,db);
			break;
			
		case UDM_URL_ACTION_FINDBYURL:
			res=UdmFindURL(A,D,db);
			break;
			
		case UDM_URL_ACTION_FINDBYMSG:
			res=UdmFindMessage(A,D,db);
			break;
			
		case UDM_URL_ACTION_FINDORIG:
			res=UdmFindOrigin(A,D,db);
			break;
			
		case UDM_URL_ACTION_EXPIRE:
			res=UdmMarkForReindex(A,db);
			break;
			
		case UDM_URL_ACTION_REFERERS:
			res=UdmGetReferers(A,db);
			break;
		
		case UDM_URL_ACTION_DOCCOUNT:
			res=UdmGetDocCount(A,db);
			break;
		
		case UDM_URL_ACTION_HTDBGET:
			res=UdmHTDBGet(A,D,db);
			break;
	        case UDM_URL_ACTION_WRITEDATA:
			if (db->DBMode == UDM_DBMODE_CACHE) {
			  return UdmURLDataWrite(A);
			}
		        return UDM_OK;

		case UDM_URL_ACTION_LINKS_DELETE:
			res = UdmDeleteLinks(A, D, db);
			break;
			
		default:
  		        UdmLog(A, UDM_LOG_ERROR, "Unsupported URL Action SQL");
			res=UDM_ERROR;
	}
	return res;
}

int UdmResActionSQL(UDM_AGENT *Agent, UDM_RESULT *Res, int cmd,UDM_DB *db){
	switch(cmd){
		case UDM_RES_ACTION_WORDS:
			return UdmFindWords(Agent,Res,db);
		case UDM_RES_ACTION_DOCINFO:
			return UdmResAddDocInfo(Agent,Res,db);
		case UDM_RES_ACTION_TARGETS:
			return UdmTargets(Agent,Res,db);
		case UDM_RES_ACTION_INSWORDS:
			return UdmResStoreBlob(Agent,Res,db);
		default:
  		        UdmLog(Agent, UDM_LOG_ERROR, "Unsupported Res Action SQL");
			return UDM_ERROR;
	}
}

int UdmCatActionSQL(UDM_AGENT *Agent, UDM_CATEGORY *Cat, int cmd,UDM_DB *db){
	switch(cmd){
		case UDM_CAT_ACTION_LIST:
			return UdmCatList(Agent,Cat,db);
		case UDM_CAT_ACTION_PATH:
			return UdmCatPath(Agent,Cat,db);
		default:
  		        UdmLog(Agent, UDM_LOG_ERROR, "Unsupported Cat Action SQL");
			return UDM_ERROR;
	}
}

int UdmSrvActionSQL(UDM_AGENT *A, UDM_SERVERLIST *S, int cmd,UDM_DB *db){
	switch(cmd){
		case UDM_SRV_ACTION_TABLE:
			return UdmLoadServerTable(A,S,db);
	        case UDM_SRV_ACTION_FLUSH:
		        return UdmServerTableFlush(db);
	        case UDM_SRV_ACTION_ADD:
		        return UdmServerTableAdd(S, db);
	        case UDM_SRV_ACTION_ID:
		        return UdmServerTableGetId(A, S, db);
   	        case UDM_SRV_ACTION_POPRANK:
		        return UdmPopRankCalculate(A, db);
		default:
  		        UdmLog(A, UDM_LOG_ERROR, "Unsupported Srv Action SQL");
			return UDM_ERROR;
	}
}

#endif /* HAVE_SQL */
